import logging
from far2l.plugin import PluginBase
from far2l.fardialogbuilder import (
    Spacer,
    TEXT,
    EDIT,
    PASSWORD,
    MASKED,
    MEMOEDIT,
    BUTTON,
    CHECKBOX,
    RADIOBUTTON,
    COMBOBOX,
    LISTBOX,
    USERCONTROL,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)

log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python Dialog Demo"
    openFrom = ["PLUGINSMENU", "COMMANDLINE", "EDITOR", "VIEWER"]

    def OpenPlugin(self, OpenFrom):
        msg2name = {}
        for name in dir(self.ffic):
            if name[:3] == 'DN_':
                msg2name[getattr(self.ffic, name)] = name
        SilentMsg = (
            self.ffic.DN_ENTERIDLE,
            self.ffic.DN_KILLFOCUS,
            self.ffic.DN_GOTFOCUS,
            self.ffic.DN_DRAWDIALOG,
            self.ffic.DN_CTLCOLORDIALOG,
            self.ffic.DN_DRAWDLGITEM,
            self.ffic.DN_DRAWDIALOGDONE,
            self.ffic.DN_CTLCOLORDLGITEM,
            self.ffic.DN_CTLCOLORDLGLIST
        )

        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg not in SilentMsg:
                log.debug(f'dlg: msg={Msg}/{msg2name.get(Msg)} Param1={Param1} Param2={Param2}')
            if Msg == self.ffic.DN_INITDIALOG:
                try:
                    dlg.SetText(dlg.ID_vapath, "vapath initial")
                    dlg.SetText(dlg.ID_vbpath, "vbpath initial")
                    dlg.Disable(dlg.ID_vbpath)
                    dlg.SetCheck(dlg.ID_vallow, 1)
                    dlg.SetFocus(dlg.ID_vseconds)
                    # override value from constructor
                    dlg.SetCheck(dlg.ID_vr1, self.ffic.BSTATE_CHECKED)
                    dlg.SetCheck(dlg.ID_vc3, self.ffic.BSTATE_CHECKED)
                except:
                    log.exception("bang")
            elif Msg == self.ffic.DN_BTNCLICK:
                pass
            elif Msg == self.ffic.DN_KEY:
                if Param2 == self.ffic.KEY_LEFT:
                    pass
                elif Param2 == self.ffic.KEY_UP:
                    pass
                elif Param2 == self.ffic.KEY_RIGHT:
                    pass
                elif Param2 == self.ffic.KEY_DOWN:
                    pass
                elif Param2 == self.ffic.KEY_ENTER:
                    pass
                elif Param2 == self.ffic.KEY_ESC:
                    pass
            elif Msg == self.ffic.DN_MOUSECLICK:
                if Param1 == dlg.ID_vlist:
                    return 1
                elif Param1 == dlg.ID_vcombo:
                    return 1
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        @self.ffi.callback("FARWINDOWPROC")
        def _DialogProc(hDlg, Msg, Param1, Param2):
            try:
                return DialogProc(hDlg, Msg, Param1, Param2)
            except:
                log.exception('dialogproc')
                return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = DialogBuilder(
            self,
            _DialogProc,
            "Python dialog",
            "helptopic",
            0,
            VSizer(
                HSizer(
                    TEXT(None, "a path"),
                    EDIT("vapath", 33, maxlength=40),
                    #                    TEXT(None, "X"),
                ),
                HSizer(
                    TEXT(None, "b path"),
                    EDIT("vbpath", 20),
                    #                    TEXT(None, "X"),
                ),
                HLine(),
                HSizer(
                    CHECKBOX("vallow", "Allow"),
                    TEXT(None, "Password"),
                    PASSWORD("vuserpass", 8, maxlength=15),
                    TEXT(None, "Seconds"),
                    MASKED("vseconds", "9999"),
                    #                    TEXT(None, "X"),
                ),
                # MEMOEDIT("vmemo", 40, 5, 512),
                HLine(),
                HSizer(
                    RADIOBUTTON("vr1", "rb1", flags=self.ffic.DIF_GROUP),
                    RADIOBUTTON("vr2", "rb2"),
                    RADIOBUTTON("vr3", "rb3", flags=self.ffic.BSTATE_CHECKED),
                    #                    TEXT(None, "X"),
                ),
                HSizer(
                    LISTBOX(
                        "vlist", 1, ["listbox A", "listbox B", "listbox C", "listbox D"]
                    ),
                    COMBOBOX(
                        "vcombo", 2, ["combo A", "combo B", "combo C", "combo D"]
                    ),
                    VSizer(
                        CHECKBOX("vc1", "cb1"),
                        CHECKBOX("vc2", "cb2"),
                        CHECKBOX("vc3", "cb3", flags=self.ffic.BSTATE_CHECKED),
                    ),
                    #                    TEXT(None, "X"),
                ),
                HLine(),
                HSizer(
                    BUTTON("vok", "OK", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vcancel", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        log.debug(
            """\
ok={}
a path=[{}] \
b path=[{}] \
allow={} \
pass=[{}] \
seconds=[{}] \
radio1={} radio2={} radio3={} \
checkbox1={} checkbox2={} checkbox3={} \
vlist={} \
vcombo={}
{}={} {}={} {}={} {}={} {}={} {}={} {}={} {}={} {}={} {}={} {}={} {}={} {}={} {}={} {}={}
""".format(
                res == dlg.ID_vok,
                dlg.GetText(dlg.ID_vapath),
                dlg.GetText(dlg.ID_vbpath),
                dlg.GetCheck(dlg.ID_vallow),
                dlg.GetText(dlg.ID_vuserpass),
                dlg.GetText(dlg.ID_vseconds),
                dlg.GetCheck(dlg.ID_vr1),
                dlg.GetCheck(dlg.ID_vr2),
                dlg.GetCheck(dlg.ID_vr3),
                dlg.GetCheck(dlg.ID_vc1),
                dlg.GetCheck(dlg.ID_vc2),
                dlg.GetCheck(dlg.ID_vc3),
                dlg.GetCurPos(dlg.ID_vlist),
                dlg.GetCurPos(dlg.ID_vcombo),
                "dlg.ID_vapath", dlg.ID_vapath,
                "dlg.ID_vbpath", dlg.ID_vbpath,
                "dlg.ID_vallow", dlg.ID_vallow,
                "dlg.ID_vuserpass", dlg.ID_vuserpass,
                "dlg.ID_vseconds", dlg.ID_vseconds,
                "dlg.ID_vr1", dlg.ID_vr1,
                "dlg.ID_vr2", dlg.ID_vr2,
                "dlg.ID_vr3", dlg.ID_vr3,
                "dlg.ID_vc1", dlg.ID_vc1,
                "dlg.ID_vc2", dlg.ID_vc2,
                "dlg.ID_vc3", dlg.ID_vc3,
                "dlg.ID_vlist", dlg.ID_vlist,
                "dlg.ID_vcombo", dlg.ID_vcombo,
                "dlg.ID_vok", dlg.ID_vok,
                "dlg.ID_vcancel", dlg.ID_vcancel,
            )
        )
        self.info.DialogFree(dlg.hDlg)



from requests import get
from time import sleep
from urllib.parse import quote
from json import loads

all_params = {"dt": ["t", "bd", "ex", "ld", "md", "qca", "rw", "rm", "ss", "t", "at"], "dj": "1", "source": "bubble"}
general_params = {"dt": ["t", "bd", "qca","t", "at",'ex','md','rw'],"dj": "1"}
# dt parameter:
#   t  - translation of source text
#   at - alternate translations
#   rm - transcription / transliteration of source and translated texts
#   bd - dictionary, in case source text is one word (you get translations with articles, reverse translations, etc.)
#   md - definitions of source text, if it's one word
#   ss - synonyms of source text, if it's one word
#   ex - examples
#   rw - See also list.
# dj - Json response with names. (dj=1)

class GoogleTranslator:
    def __init__(self,name,base_url,general_params):
        """Google Translator can support multiple source."""
        self.name=name
        self.url = base_url
        self.base_parames=general_params
        self.client_list={
            'nottk':['gtx','at'],
            'ttk':['webapp'],
            'clients5':['dict-chrome-ex']
        }
        self.client_list_state=self.client_list.copy()

    @staticmethod
    def convert_qtext(text):
        return quote(text)

    @staticmethod
    def calculate_tkid(text):
        return tk.calculate_token(text)

    def __call__(self,input_text, sourcelanguage='auto',targetlanguage='tr',ttk_enable=False):
        HEADERS = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36",
            "Accept": "*/*",
            "Accept-Language": "en-US,en-GB; q=0.5",
            "Accept-Encoding": "gzip, deflate",
            "Content-Type": "application/x-www-form-urlencoded; application/json; charset=UTF-8",
            "Connection": "keep-alive"
        }
        result_dict={}
        #text = self.convert_qtext(input_text)
        self.parames=general_params
        self.parames['sl']=sourcelanguage
        self.parames['tl']=targetlanguage
        self.parames['hl']=targetlanguage
        self.parames['q']=input_text
        self.parames['ie']='UTF-8'
        self.parames['oe']='UTF-8'
        mode=''
        if self.name == 'clients5' and not ttk_enable:
            mode='clients5'
        elif ttk_enable:
            mode = 'ttk'
            tkid = self.calculate_tkid(input_text)
            self.parames['tk']=tkid
        else:
            mode='nottk'
        #print(mode,self.client_list[mode])
        self.parames['client']=self.client_list[mode][0]

        try:
            req = get(self.url,params=self.parames,headers=HEADERS)
            status_code = req.status_code
        except Exception :
            return 400,None

        if status_code == 400:
            return status_code,None

        while  status_code != 200:
            self.client_list[mode].pop(0)
            if self.client_list[mode] == []:
                return False,None
            self.parames['client']=self.client_list[mode][0]
            req = get(self.url,params=self.parames,headers=HEADERS)
            status_code = req.status_code

        return status_code,self.parser(req.text)


    def parser(self,request_text):
        req = loads(request_text)
        if 'definitions' in req.keys():
            definitions={}
            for p in  req['definitions']:
                definitions[p['pos']]= [ {'detail':g['gloss'],'example':g['example'] if 'example' in g.keys() else None} for g in p['entry']]
        else:
            definitions =None
        if 'examples' in req.keys():
            examples=[e['text'].replace('<b>','').replace('</b>','') for e in req['examples']['example']]
        else:
            examples=None
        detect_language = req['ld_result']['extended_srclangs'][0]#req['src']

        if self.name  == 'clients5':
            req['sentences']=req['sentences'][:-1]
        result = [ trans['trans'] for trans in req['sentences']]

        if 'spell' in req and 'spell_res' in req['spell']:
            revise=req['spell']['spell_res']
        else:
            revise=None

        if 'dict' in req:
            dict_list = req['dict']
            allresult = [{'pos': dic['pos'],'terms':dic['terms'] } for dic in dict_list]
        else:
            allresult=[]
        result_dict = {}
        result_dict['result']=result
        result_dict['all_result']=allresult
        result_dict['detect_language']=detect_language
        result_dict['revise']=revise
        result_dict['definition']=definitions
        result_dict['example']=examples
        return result_dict

#service
service_dict={
    'com_apis':'https://translate.googleapis.com/translate_a/single',
    'clients5':'https://clients5.google.com/translate_a/t',
    'com':'https://translate.google.com/translate_a/single',
    'tw':'https://translate.google.com.tw/translate_a/single',
    'cn':'https://translate.google.cn/translate_a/single',
}

translator_list = [GoogleTranslator(name,url,general_params) for name,url in service_dict.items()]

translator_idx=0
ttk_enable=False

RETRY_TIME=60*5

def get_translate(inputtext, sourcelanguage='auto',targetlanguage='zh-TW'):
    global ttk_enable,translator_idx,RETRY_TIME

    status,result_dict=translator_list[translator_idx](inputtext, sourcelanguage,targetlanguage,ttk_enable)

    if status == 200:
        return result_dict
    elif status == 400:
        return None
    if True:
        return None
    else:
        translator_idx +=1
        if translator_idx < len(translator_list):
            return get_translate(inputtext, sourcelanguage,targetlanguage)
        elif not ttk_enable:
            translator_idx=0
            ttk_enable=True
            return get_translate(inputtext, sourcelanguage,targetlanguage)
        else:
            sleep(RETRY_TIME)
            translator_idx=0
            ttk_enable=False
            return get_translate(inputtext, sourcelanguage,targetlanguage)

if __name__ == '__main__':
    print(get_translate("And do beautifull things", "en", "pl-PL"))



import logging
from far2l.plugin import PluginBase
from far2l.fardialogbuilder import (
    Spacer,
    TEXT,
    EDIT,
    PASSWORD,
    MASKED,
    MEMOEDIT,
    BUTTON,
    CHECKBOX,
    RADIOBUTTON,
    COMBOBOX,
    LISTBOX,
    USERCONTROL,
    HLine,
    HSizer,
    VSizer,
    FlowSizer,
    DialogBuilder,
)

log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python Dialog Demo"
    openFrom = ["PLUGINSMENU", "COMMANDLINE", "EDITOR", "VIEWER"]

    def OpenPlugin(self, OpenFrom):
        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                try:
                    dlg.SetText(dlg.ID_vapath, "vapath initial")
                    dlg.SetText(dlg.ID_vbpath, "vbpath initial")
                    dlg.Disable(dlg.ID_vbpath)
                    dlg.SetCheck(dlg.ID_vallow, 1)
                    dlg.SetFocus(dlg.ID_vseconds)
                    # override value from constructor
                    dlg.SetCheck(dlg.ID_vr1, self.ffic.BSTATE_CHECKED)
                    dlg.SetCheck(dlg.ID_vc3, self.ffic.BSTATE_CHECKED)
                except:
                    log.exception("bang")
            elif Msg == self.ffic.DN_BTNCLICK:
                pass
            elif Msg == self.ffic.DN_KEY:
                if Param2 == self.ffic.KEY_LEFT:
                    pass
                elif Param2 == self.ffic.KEY_UP:
                    pass
                elif Param2 == self.ffic.KEY_RIGHT:
                    pass
                elif Param2 == self.ffic.KEY_DOWN:
                    pass
                elif Param2 == self.ffic.KEY_ENTER:
                    pass
                elif Param2 == self.ffic.KEY_ESC:
                    pass
            elif Msg == self.ffic.DN_MOUSECLICK:
                pass
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = DialogBuilder(
            self,
            DialogProc,
            "Python dialog",
            "helptopic",
            0,
            VSizer(
                FlowSizer(3,
                    HSizer(
                        TEXT(None, "a path"),
                        EDIT("vapath", 33, maxlength=40),
                    ),
                    HSizer(
                        TEXT(None, "b path"),
                        EDIT("vbpath", 20),
                    ),
                    CHECKBOX("vallow", "Allow"),
                    HSizer(
                        TEXT(None, "Password"),
                        PASSWORD("vuserpass", 8, maxlength=15),
                    ),
                    HSizer(
                        TEXT(None, "Seconds"),
                        MASKED("vseconds", "9999"),
                    ),
                    VSizer(
                        RADIOBUTTON("vr1", "p1", flags=self.ffic.DIF_GROUP),
                        RADIOBUTTON("vr2", "p2"),
                        RADIOBUTTON("vr3", "p3", flags=self.ffic.BSTATE_CHECKED),
                    ),
                    LISTBOX(
                        "vlist", 1, ["element A", "element B", "element C", "element D"]
                    ),
                    COMBOBOX(
                        "vcombo", 2, ["element A", "element B", "element C", "element D"]
                    ),
                    VSizer(
                        CHECKBOX("vc1", "c1"),
                        CHECKBOX("vc2", "c2"),
                        CHECKBOX("vc3", "c3", flags=self.ffic.BSTATE_CHECKED),
                    ),
                    #                    TEXT(None, "X"),
                ),
                HLine(),
                HSizer(
                    BUTTON("vok", "OK", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vcancel", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        log.debug(
            """\
ok={} \
a path=[{}] \
b path=[{}] \
allow={} \
pass=[{}] \
seconds=[{}] \
radio1={} \
radio2={} \
radio3={} \
checkbox1={} \
checkbox2={} \
checkbox3={} \
vlist={} \
vcombo={} \
""".format(
                res == dlg.ID_vok,
                dlg.GetText(dlg.ID_vapath),
                dlg.GetText(dlg.ID_vbpath),
                dlg.GetCheck(dlg.ID_vallow),
                dlg.GetText(dlg.ID_vuserpass),
                dlg.GetText(dlg.ID_vseconds),
                dlg.GetCheck(dlg.ID_vr1),
                dlg.GetCheck(dlg.ID_vr2),
                dlg.GetCheck(dlg.ID_vr3),
                dlg.GetCheck(dlg.ID_vc1),
                dlg.GetCheck(dlg.ID_vc2),
                dlg.GetCheck(dlg.ID_vc3),
                dlg.GetCurPos(dlg.ID_vlist),
                dlg.GetCurPos(dlg.ID_vcombo),
            )
        )
        self.info.DialogFree(dlg.hDlg)



class Editor:
    def __init__(self, plugin):
        self.info = plugin.info
        self.ffic = plugin.ffic
        self.ffi = plugin.ffi
        self.s2f = plugin.s2f
        self.f2s = plugin.f2s

    def GetFileName(self):
        nb = self.info.EditorControl(self.ffic.ECTL_GETFILENAME, self.ffi.NULL)
        if nb:
            data = self.ffi.new("wchar_t []", nb)
            self.info.EditorControl(self.ffic.ECTL_GETFILENAME, self.ffi.cast("PVOID", data))
            return self.f2s(data)
        return None



import threading
from far2l.plugin import PluginBase

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


import logging
log = logging.getLogger(__name__)

class MyWindow(Gtk.Window):
    def __init__(self):
        super().__init__(title="Hello World")

        self.button = Gtk.Button(label="Click Here")
        self.button.connect("clicked", self.on_button_clicked)
        self.add(self.button)

        self.connect("destroy", self.onQuit)
        self.show_all()

    def on_button_clicked(self, widget):
        log.debug("Hello World")

    def onQuit(self, window):
        Gtk.main_quit()

class Plugin(PluginBase):
    label = "Python GTK"
    openFrom = ["PLUGINSMENU", "EDITOR"]

    def OpenPlugin(self, OpenFrom):
        if OpenFrom == 5:
            # EDITOR
            def proc():
                win = MyWindow()
                Gtk.main()
            t = threading.Thread(target=proc)
            t.daemon = True
            t.start()
        return -1



import os
import logging

import ctypes as ct
import far2lc
from far2l.plugin import PluginBase

log = logging.getLogger(__name__)

CAH_ALLOCATED_MAGIC = 0x0610ba10A110CED0
CAH_FREED_MAGIC     = 0x0610ba10F4EED000
class ClipboardAllocHeader(ct.Structure):
    _fields_ = [
		('size', ct.c_uint32),
		('padding', ct.c_uint32),
		('magic', ct.c_uint64),
    ]

class Plugin(PluginBase):
    label = "Python Clip SET"
    openFrom = ["PLUGINSMENU", 'FILEPANEL']

    def SetClipboard(self, winport, fmt, fqname):
        fqname = fqname.encode('utf8')
        #log.debug(f'clip.{fmt} set: {fqname} {len(fqname)}')
        ptr = winport.ClipboardAlloc(len(fqname))
        if ptr is None:
            log.debug('clipboard alloc failed')
        else:
            if 0:
                m = int(self.ffi.cast('DWORD64', ptr))
                log.debug(f'ptr={ptr} m={m:x}')
                m -= ct.sizeof(ClipboardAllocHeader)
                log.debug(f'm={m:x}')
                mp = ClipboardAllocHeader.from_address(m)
                log.debug(f'mp.size={mp.size} mp.padding={mp.padding} mp.magic={mp.magic:x} CAH_ALLOCATED_MAGIC={CAH_ALLOCATED_MAGIC:x}')
            self.ffi.memmove(ptr, fqname, len(fqname)) 
            res = winport.SetClipboardData(fmt, ptr)

    def OpenPlugin(self, OpenFrom):
        winport = self.ffi.cast("struct WINPORTDECL *", far2lc.WINPORT())
        clipurifmt = winport.RegisterClipboardFormat("text/uri-list")
        if not clipurifmt:
            log.error('uclipset.ClipboardRegisterFormat.1')
            return
        clipgnofmt = winport.RegisterClipboardFormat("x-special/gnome-copied-files")
        if not clipgnofmt:
            log.error('uclipset.ClipboardRegisterFormat.2')
            return
        if not winport.OpenClipboard(self.ffi.NULL):
            log.error('uclipset.OpenClipboard')
            return
        try:
            files = []
            pwd = self.panel.GetPanelDir()
            for i in range(self.panel.GetPanelInfo().SelectedItemsNumber):
                pnli, pnlidata = self.panel.GetSelectedPanelItem(i)
                fqname = self.f2s(pnli.FindData.lpwszFileName)
                files.append("file://"+os.path.join(pwd, fqname))
            #self.SetClipboard(winport, 1, fqname)
            self.SetClipboard(winport, clipurifmt, "\n".join(files))
            self.SetClipboard(winport, clipgnofmt, "copy\n"+"\n".join(files))
        except:
            log.exception('uclipset.SetClipboard')
        finally:
            if not winport.CloseClipboard():
                log.error('uclipset.CloseClipboard')
                return
        # typedef BOOL (*WINPORT_IsClipboardFormatAvailable) (UINT format);
        # typedef PVOID (*WINPORT_GetClipboardData) (UINT format);
        # typedef SIZE_T (*WINPORT_ClipboardSize) (PVOID mem);



import logging
from .far2lpanel import Panel


log = logging.getLogger(__name__)


class PluginBase:
    # private
    name = ""
    number = 0
    # public
    USERHOME = ""  # ~/.config/far2l/plugins/python
    label = ""
    # openFrom = ["DISKMENU", "PLUGINSMENU", "FINDLIST", "SHORTCUT", "COMMANDLINE", "EDITOR", "VIEWER", "FILEPANEL"]
    openFrom = []
    Configure = None  # override with method when configuration dialog is needed

    def __init__(self, parent, info, ffi, ffic):
        self.parent = parent
        self.info = info
        self.ffi = ffi
        self.ffic = ffic
        self.hplugin = self.ffi.cast("void *", id(self))
        self.panel = Panel(self)

    def s2f(self, s):
        return self.ffi.new("wchar_t []", s)

    def f2s(self, s):
        return self.ffi.string(self.ffi.cast("wchar_t *", s))

    def notice(self, body, title=None):
        return self._popup(body, title, 0)

    def error(self, body, title=None):
        return self._popup(body, title, self.ffic.FMSG_WARNING)

    def _popup(self, body, title, flags):
        items = [self.s2f(title)] if title else [self.s2f('')]
        if isinstance(body, str):
            if '\n' in body:
                body = body.split('\n')
            else:
                items.append(self.s2f(body))
        if not isinstance(body, str):
            items.extend(self.s2f(s) for s in body)
        # note: needs to be at least 2 items, otherwise message 
        # box is not shown
        citems = self.ffi.new('wchar_t *[]', items)
        if not flags:
            flags = self.ffic.FMSG_MB_OK
        return self.info.Message(
            self.info.ModuleNumber,  # GUID
            flags,  # Flags
            self.ffi.NULL,  # HelpTopic
            citems, len(citems), 1,  # ButtonsNumber
        )

    @staticmethod
    def HandleCommandLine(line):
        log.debug("Plugin.HandleCommandLine({})".format(line))
        return False

    def OpenPlugin(self, OpenFrom):
        log.debug("Plugin.OpenPlugin({}) #{}".format(OpenFrom, self.label))

    def Close(self):
        log.debug("Plugin.Close() #{}".format(self.label))

    def GetOpenPluginInfo(self, OpenInfo):
        log.debug("Plugin.GetOpenPluginInfo({}) #{}".format(OpenInfo, self.label))

    def ProcessKey(self, Key, ControlState):
        # log.debug("Plugin.ProcessKey({0}, {1})".format(Key, ControlState))
        return 0

    def MayExitFAR(self):
        # log.debug("Plugin.MayExitFAR()")
        return True

    def ExitFAR(self):
        #log.debug("Plugin.ExitFAR()")
        pass

    @staticmethod
    def OpenFilePlugin(parent, info, ffi, ffic, Name, Data, DataSize, OpMode):
        return False

    def ProcessDialogEvent(self, Event, Param):
        # log.debug("Plugin.ProcessDialogEvent({0}, {1})".format(Event, Param))
        return 0

    def ProcessEditorEvent(self, Event, Param):
        # log.debug("Plugin.ProcessEditorEvent({0}, {1})".format(Event, Param))
        return 0

    def ProcessEditorInput(self, Rec):
        # log.debug("Plugin.ProcessEditorInput({0})".format(Rec))
        return 0

    def ProcessSynchroEvent(self, Event, Param):
        # log.debug("Plugin.ProcessSynchroEvent({0}, {1})".format(Event, Param))
        return 0

    def ProcessViewerEvent(self, Event, Param):
        # log.debug("Plugin.ProcessViewerEvent({0}, {1})".format(Event, Param))
        return 0


class PluginVFS(PluginBase):
    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        log.debug(
            "VFS.GetFindData({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode)
        )
        return 0

    def FreeFindData(self, PanelItem, ItemsNumber):
        log.debug("VFS.FreeFindData({0}, {1})".format(PanelItem, ItemsNumber))

    def GetVirtualFindData(self, PanelItem, ItemsNumber, Path):
        log.debug(
            "VFS.GetVirtualFindData({0}, {1}, {2})".format(PanelItem, ItemsNumber, Path)
        )
        return 0

    def FreeVirtualFindData(self, PanelItem, ItemsNumber):
        log.debug("VFS.FreeVirtualFindData({0}, {1})".format(PanelItem, ItemsNumber))

    def Compare(self, PanelItem1, PanelItem2, Mode):
        log.debug("VFS.Compare({0}, {1}, {2})".format(PanelItem1, PanelItem2, Mode))
        return -2

    def GetFiles(self, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        log.debug(
            "VFS.GetFiles({0}, {1}, {2}, {3}, {4})".format(
                PanelItem,
                ItemsNumber,
                Move,
                DestPath,
                OpMode,
            )
        )
        return 0

    def PutFiles(self, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        log.debug(
            "VFS.PutFiles({0}, {1}, {2}, {3}, {4})".format(
                PanelItem,
                ItemsNumber,
                Move,
                SrcPath,
                OpMode,
            )
        )
        return 0

    def MakeDirectory(self, Name, OpMode):
        log.debug("VFS.MakeDirectory({0}, {1})".format(Name, OpMode))
        return 0

    def SetDirectory(self, Dir, OpMode):
        log.debug("VFS.SetDirectory({0}, {1})".format(Dir, OpMode))
        return 0

    def DeleteFiles(self, PanelItem, ItemsNumber, OpMode):
        log.debug(
            "VFS.DeleteFiles({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode)
        )
        return 0

    def SetFindList(self, PanelItem, ItemsNumber):
        log.debug("VFS.SetFindList({0}, {1})".format(PanelItem, ItemsNumber))
        return 0

    def ProcessEvent(self, Event, Param):
        log.debug("VFS.ProcessEvent({0}, {1})".format(Event, Param))
        return 0

    def ProcessHostFile(self, PanelItem, ItemsNumber, OpMode):
        log.debug(
            "VFS.ProcessHostFile({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode)
        )
        return 0



import time
import logging
import threading
from far2l.plugin import PluginBase
from far2l.fardialogbuilder import (
    Spacer,
    TEXT,
    BUTTON,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)

log = logging.getLogger(__name__)


class ProgressThread(threading.Thread):
    def __init__(self):
        super().__init__()
        self.cond = threading.Lock()
        self.daemon = True
        self.sleep = 0.1
        self.vmin = 0
        self.vmax = 100
        self.done = False

    def cancel(self):
        if not self.done:
            self.vmin = -self.vmin

    def run(self):
        while self.vmin >= 0 and self.vmin < self.vmax:
            self.cond.acquire(True)
            self.vmin += 1
            self.cond.release()
            time.sleep(self.sleep)
        self.done = True

class Plugin(PluginBase):
    label = "Python Progress Dialog"
    openFrom = ["PLUGINSMENU"]

    def OnIdle(self, t, dlg):
        if not t.cond.acquire(False):
            return
        self._OnIdle(t, dlg)
        t.cond.release()

    def _OnIdle(self, t, dlg):
        n = t.vmin * 20 // t.vmax
        fprog = (('='*n)+' '*20)[:20]
        dlg.SetText(dlg.ID_fprog, '[{}]'.format(fprog))

    def OpenPlugin(self, OpenFrom):
        t = ProgressThread()

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                dlg.SetText(dlg.ID_fprog, "")
                # start thread
                t.start()
            elif Msg == self.ffic.DN_ENTERIDLE:
                if t.done:
                    dlg.Close(0)
                else:
                    self.OnIdle(t, dlg)
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = DialogBuilder(
            self,
            DialogProc,
            "Python Progress Demo",
            "helptopic",
            self.ffic.FDLG_REGULARIDLE,
            VSizer(
                TEXT("fprog", " "*(2+20)),
                HLine(),
                BUTTON("vcancel", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)

        t.cancel()
        v = t.vmin
        del t
        log.debug('t.vmin={} res={}'.format(v, res))

        self.info.DialogFree(dlg.hDlg)



import io
import logging
from . import fardialogsizer as sizer
from .fardialog import Dialog


log = logging.getLogger(__name__)


class Element(sizer.Window):
    no = 0
    varname = None

    def __init__(
        self, varname=None, focus=0, param=None, flags=0, default=0, maxlength=0
    ):
        self.varname = varname
        self.focus = focus
        self.param = param or {"Selected": 0}
        self.flags = flags
        self.default = default
        self.maxlength = maxlength
        self.data = None

    def makeID(self, dlg):
        self.no = Element.no
        Element.no += 1
        if self.varname is not None:
            setattr(dlg, "ID_" + self.varname, self.no)

    def makeItem(self, dlg):
        w, h = self.get_best_size()
        if self.data is None:
            self.data = dlg.ffi.NULL
        item = [
            getattr(dlg.ffic, self.dit),
            self.pos[0],
            self.pos[1],
            self.pos[0] + w - 1,
            self.pos[1] + h - 1,
            self.focus,
            self.param,
            self.flags,
            self.default,
            self.data,
            self.maxlength,
        ]
        dlg.dialogItems.append(item)
        # if self.varname is not None:
        #    dlg.id2item[getattr(dlg, "ID_"+self.varname)] = [item, None]


class Spacer(Element):
    def __init__(self, width=1, height=1):
        super().__init__()
        self.width = width
        self.height = height

    def makeID(self, dlg):
        # non countable ID
        pass

    def get_best_size(self):
        return (self.width, self.height)

    def makeItem(self, dlg):
        pass

class TEXT(Element):
    dit = "DI_TEXT"

    def __init__(self, varname, text, **kwargs):
        super().__init__(varname, **kwargs)
        self.text = text

    def get_best_size(self):
        return (len(self.text.replace('&','')), 1)

    def makeItem(self, dlg):
        self.data = dlg.s2f(self.text)
        super().makeItem(dlg)


class EDIT(Element):
    dit = "DI_EDIT"

    def __init__(self, varname, width, mask=None, height=1, **kwargs):
        if "history" in kwargs:
            history = kwargs.pop("history")
        else:
            history = None
        super().__init__(varname, **kwargs)
        self.width = width
        self.height = height
        self.mask = None
        self.history = history

    def get_best_size(self):
        return (self.width, self.height)

    def makeItem(self, dlg):
        self.data = dlg.ffi.NULL
        if self.mask is not None:
            self.param = {"Mask": dlg.s2f(self.mask)}
            self.flags = dlg.ffic.DIF_MASKEDIT
        elif self.history is not None:
            self.param = {"History": dlg.s2f(self.history)}
            self.flags = dlg.ffic.DIF_HISTORY
        super().makeItem(dlg)


class PASSWORD(EDIT):
    dit = "DI_PSWEDIT"


class MASKED(EDIT):
    dit = "DI_FIXEDIT"

    def __init__(self, varname, mask, **kwargs):
        super().__init__(varname, len(mask), mask=mask, maxlength=len(mask), **kwargs)


class MEMOEDIT(EDIT):
    # dit = "DI_MEMOEDIT"

    def __init__(self, varname, width, height, maxlength=None):
        raise NotImplemented("MEMOEDIT")


class BUTTON(Element):
    dit = "DI_BUTTON"

    def __init__(self, varname, text, **kwargs):
        super().__init__(varname, **kwargs)
        self.text = text

    def get_best_size(self):
        # [ text ]
        return (4 + len(self.text.replace('&','')), 1)

    def makeItem(self, dlg):
        self.data = dlg.s2f(self.text)
        self.default = dlg.ffic.DIF_DEFAULT if self.default else 0
        super().makeItem(dlg)


class CHECKBOX(Element):
    dit = "DI_CHECKBOX"

    def __init__(self, varname, text, checked=0, **kwargs):
        super().__init__(varname, **kwargs)
        self.text = text
        self.param = {"Selected": checked}

    def get_best_size(self):
        # [?] text
        return (4 + len(self.text.replace('&','')), 1)

    def makeItem(self, dlg):
        self.data = dlg.s2f(self.text)
        super().makeItem(dlg)


class RADIOBUTTON(Element):
    dit = "DI_RADIOBUTTON"

    def __init__(self, varname, text, selected=0, **kwargs):
        super().__init__(varname, **kwargs)
        self.text = text
        self.param = {"Selected": selected}

    def get_best_size(self):
        # (?) text
        return (4 + len(self.text.replace('&','')), 1)

    def makeItem(self, dlg):
        self.data = dlg.s2f(self.text)
        super().makeItem(dlg)


class COMBOBOX(Element):
    dit = "DI_COMBOBOX"

    def __init__(self, varname, selected, items, **kwargs):
        if "width" in kwargs:
            width = kwargs.pop("width")
        else:
            width = max([len(s) for s in items])
        super().__init__(varname, **kwargs)
        self.selected = selected
        self.items = items
        self.width = width

    def get_best_size(self):
        return (2 + self.width, 1)

    def makeItem(self, dlg):
        self.data = dlg.ffi.NULL
        s2f = []
        fitems = dlg.ffi.new("struct FarListItem []", len(self.items))
        for i in range(len(self.items)):
            fitems[i].Flags = dlg.ffic.LIF_SELECTED if i == self.selected else 0
            s = dlg.s2f(self.items[i])
            s2f.append(s)
            fitems[i].Text = s

        flist = dlg.ffi.new("struct FarList *")
        flist.ItemsNumber = len(self.items)
        flist.Items = fitems

        self.flist = flist
        self.fitems = fitems
        self.s2f = s2f

        self.param = {"ListItems": flist}
        super().makeItem(dlg)


class LISTBOX(Element):
    dit = "DI_LISTBOX"

    def __init__(self, varname, selected, items, **kwargs):
        if "height" in kwargs:
            height = kwargs.pop("height")
        else:
            height = len(items) + 2
        super().__init__(varname, **kwargs)
        self.selected = selected
        self.items = items
        self.width = max([len(s) for s in self.items])
        self.height = height

    def get_best_size(self):
        return (4 + self.width, self.height)

    def makeItem(self, dlg):
        self.data = dlg.ffi.NULL
        s2f = []
        fitems = dlg.ffi.new("struct FarListItem []", len(self.items))
        for i in range(len(self.items)):
            fitems[i].Flags = dlg.ffic.LIF_SELECTED if i == self.selected else 0
            s = dlg.s2f(self.items[i])
            s2f.append(s)
            fitems[i].Text = s

        flist = dlg.ffi.new("struct FarList *")
        flist.ItemsNumber = len(self.items)
        flist.Items = fitems

        self.flist = flist
        self.fitems = fitems
        self.s2f = s2f

        self.param = {"ListItems": flist}
        super().makeItem(dlg)


class USERCONTROL(Element):
    dit = "DI_USERCONTROL"

    def __init__(self, varname, width, height, **kwargs):
        super().__init__(varname, **kwargs)
        self.width = width
        self.height = height

    def get_best_size(self):
        return (self.width, self.height)

    def makeItem(self, dlg):
        self.data = dlg.ffi.NULL
        super().makeItem(dlg)


class HLine(Element):
    dit = "DI_TEXT"

    def __init__(self):
        super().__init__()

    def get_best_size(self):
        return (1, 1)

    def makeItem(self, dlg):
        self.data = dlg.ffi.NULL
        self.flags = dlg.ffic.DIF_SEPARATOR
        dlg.dialogItems.append(
            (
                getattr(dlg.ffic, self.dit),
                0,
                self.pos[1],
                0,
                self.pos[1],
                self.focus,
                self.param,
                self.flags,
                self.default,
                self.data,
                self.maxlength,
            )
        )


class HSizer(sizer.HSizer):
    def __init__(self, *controls, border=(0, 0, 1, 0), center=False):
        super().__init__()
        self.controls = controls
        self.center = center
        for control in controls:
            self.add(control, border)

    def makeID(self, dlg):
        for control in self.controls:
            control.makeID(dlg)

    def makeItem(self, dlg):
        for control in self.controls:
            control.makeItem(dlg)


class VSizer(sizer.VSizer):
    def __init__(self, *controls, border=(0, 0, 0, 0)):
        super().__init__()
        self.controls = controls
        for control in controls:
            self.add(control, border)

    def makeID(self, dlg):
        for control in self.controls:
            control.makeID(dlg)

    def makeItem(self, dlg):
        for control in self.controls:
            control.makeItem(dlg)


class FlowSizer(sizer.FlowSizer):
    def __init__(self, cols, *controls, border=(0, 0, 0, 0)):
        super().__init__(cols)
        self.controls = controls
        for control in controls:
            self.add(control, border)

    def makeID(self, dlg):
        for control in self.controls:
            control.makeID(dlg)

    def makeItem(self, dlg):
        for control in self.controls:
            control.makeItem(dlg)


class DialogBuilder(sizer.HSizer):
    def __init__(
        self, plugin, dialogProc, title, helptopic, flags, contents, border=(2, 1, 2, 1)
    ):
        super().__init__(border)
        self.plugin = plugin
        self.dialogProc = dialogProc
        self.title = title
        self.helptopic = helptopic
        self.flags = flags
        self.contents = contents
        self.add(contents, border)

    def build(self, x, y):

        dlg = Dialog(self.plugin)

        w, h = self.get_best_size()
        w = max(w + 1, len(self.title) + 2)

        dlg.width = w
        dlg.height = h

        # for building dlg.ID_<varname>
        Element.no = 1
        self.contents.makeID(dlg)
        self.size(3, 1, w, h)

        dlg.dialogItems.insert(0,
            (
                dlg.ffic.DI_DOUBLEBOX,
                3,
                1,
                w,
                h,
                0,
                {"Selected": 0},
                0,
                0,
                dlg.s2f(self.title),
                0,
            )
        )
        self.contents.makeItem(dlg)

        # for item in dlg.dialogItems:
        #    log.debug('{}'.format(item))

        dlg.fdi = dlg.ffi.new("struct FarDialogItem []", dlg.dialogItems)
        dlg.hDlg = dlg.info.DialogInit(
            dlg.info.ModuleNumber,
            x,
            y,
            w + 4,
            h + 2,
            dlg.s2f(self.helptopic),
            dlg.fdi,
            len(dlg.fdi),
            0,
            self.flags,
            self.dialogProc,
            0,
        )
        return dlg

    def build_nobox(self, x, y, w, h):

        dlg = Dialog(self.plugin)

        dlg.width = w
        dlg.height = h

        # for building dlg.ID_<varname>
        Element.no = 0
        self.contents.makeID(dlg)
        self.size(0, 0, w, h)

        self.contents.makeItem(dlg)

        dlg.fdi = dlg.ffi.new("struct FarDialogItem []", dlg.dialogItems)
        dlg.hDlg = dlg.info.DialogInit(
            dlg.info.ModuleNumber,
            x,
            y,
            w,
            h,
            dlg.s2f(self.helptopic),
            dlg.fdi,
            len(dlg.fdi),
            0,
            self.flags,
            self.dialogProc,
            0,
        )
        return dlg



import os
import time
import stat
import re
import logging
from datetime import datetime
import configparser

from far2l.plugin import PluginVFS
from far2l.fardialogbuilder import (
    Spacer,
    TEXT,
    EDIT,
    PASSWORD,
    MASKED,
    MEMOEDIT,
    BUTTON,
    CHECKBOX,
    RADIOBUTTON,
    COMBOBOX,
    LISTBOX,
    USERCONTROL,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)

from adbutils import AdbClient, AdbError

log = logging.getLogger(__name__)

TICKS_PER_SECOND = 10000000
EPOCH_DIFFERENCE = 11644473600

FILE_ATTRIBUTE_READONLY = 0x00000001
FILE_ATTRIBUTE_HIDDEN = 0x00000002
FILE_ATTRIBUTE_SYSTEM = 0x00000004
FILE_ATTRIBUTE_DIRECTORY = 0x00000010
FILE_ATTRIBUTE_ARCHIVE = 0x00000020
FILE_ATTRIBUTE_DEVICE = 0x00000040
FILE_ATTRIBUTE_NORMAL = 0x00000080
FILE_ATTRIBUTE_TEMPORARY = 0x00000100
FILE_ATTRIBUTE_SPARSE_FILE = 0x00000200
FILE_ATTRIBUTE_REPARSE_POINT = 0x00000400
FILE_ATTRIBUTE_COMPRESSED = 0x00000800
FILE_ATTRIBUTE_OFFLINE = 0x00001000
FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x00002000
FILE_ATTRIBUTE_ENCRYPTED = 0x00004000
FILE_ATTRIBUTE_INTEGRITY_STREAM = 0x00008000
FILE_ATTRIBUTE_VIRTUAL = 0x00010000
FILE_ATTRIBUTE_NO_SCRUB_DATA = 0x00020000
FILE_ATTRIBUTE_BROKEN = 0x00200000
FILE_ATTRIBUTE_EXECUTABLE = 0x00400000


class Config:
    def __init__(self, home):
        self.inipath = os.path.join(home, "adb.ini")
        self.config = configparser.ConfigParser()
        self.config["adb"] = {
            "box": "default",
            "rexec": "False",
        }
        self.load()

    def load(self):
        try:
            self._load()
        except:
            log.exception("load config")

    def _load(self):
        if os.path.isfile(self.inipath):
            with open(self.inipath, "rt") as fp:
                self.config.read_file(fp)

    def save(self):
        try:
            self._save()
        except:
            log.exception("save config")

    def _save(self):
        with open(self.inipath, "wt") as fp:
            self.config.write(fp)

    @property
    def box(self):
        return self.config.get("adb", "box")

    @box.setter
    def box(self, value):
        self.config.set("adb", "box", value)

    @property
    def rexec(self):
        return self.config.getboolean("adb", "rexec")

    @rexec.setter
    def rexec(self, value):
        self.config.set("adb", "rexec", str(value))


class Entry:
    def __init__(
        self,
        dirname=None,
        perms=None,
        links=1,
        uid=0,
        gid=0,
        size=None,
        date_time=None,
        date=None,
        name=None,
    ):
        """initialize file"""
        self.mode = 0
        self.perms = perms
        self.links = links
        self.uid = uid
        self.gid = gid
        self.size = int(size) if size else 0
        self.date_time = date_time
        self.name = name
        self.date = date

        self.dirname = dirname
        self.link_target = None
        self.filepath = os.path.join(self.dirname, self.name)
        self.perms2mode()

    def update(self, box):
        """update object fields"""
        month_num = {
            "Jan": 1,
            "Feb": 2,
            "Mar": 3,
            "Apr": 4,
            "May": 5,
            "Jun": 6,
            "Jul": 7,
            "Aug": 8,
            "Sep": 9,
            "Oct": 10,
            "Nov": 11,
            "Dec": 12,
        }
        if self.date_time:
            date = self.date_time.split()
            date = "%s-%02d-%s %s" % (date[1], month_num[date[0]], date[3], date[2])
            date = datetime.strptime(date, "%d-%m-%Y %H:%M:%S")
        else:
            date = datetime.strptime(self.date, box["date_re"])

        self.date = date
        self.date_time = date.strftime("%Y-%m-%d %H:%M")

        type = self.perms[0] if self.perms else None

        if type == "l" and " -> " in self.name:
            self._correct_link()

    def perms2mode(self):
        perms = self.perms[1:]
        lperms = len(perms)
        mode = 0
        # Check if 'perms' has the right format
        if not [x for x in perms if x not in "-rwxXsStT"] and lperms == 9:
            pos = lperms - 1
            for c in perms:
                if c in "sStT":
                    # Special modes
                    mode += (1 << pos // 3) << 9

                mode += 1 << pos if c in "rwxst" else 0
                pos -= 1
        self.mode = (
            mode
            | {
                "-": 0,
                "b": stat.S_IFBLK,
                "c": stat.S_IFCHR,
                "d": stat.S_IFDIR,
                "l": stat.S_IFLNK,
                "p": 0,  # stat.S_IF???,
                "s": 0,  # stat.S_IF???,
            }[self.perms[0]]
        )

    def _correct_link(self):
        """Canonize filename and fill the link attr"""
        try:
            name, target = self.name.split(" -> ")
        except ValueError:
            return

        self.name = name

        if not self.size:
            self.size = 0

        if target.startswith("/"):
            self.link_target = target
        else:
            self.link_target = os.path.abspath(os.path.join(self.dirname, target))
        self.filepath = os.path.join(self.dirname, self.name)

    def mk_link_relative(self):
        """Convert links to relative"""
        self.link_target = os.path.relpath(self.link_target, self.dirname)

    def __str__(self):
        if not self.name:
            return ""

        template = (
            "{mode:4o} {perms} {links:>4} {uid:<8} {gid:<8} {size:>8} {date} {name}"
        )
        return template.format(**self.__dict__)


class Shell:
    boxes = {
        "busybox": {
            "ls": "busybox ls -anL --full-time {}",
            "rls": "busybox ls -Ranl {}",
            "date_re": "%Y-%m-%d %H:%M:%S %z",
            "file_re": r"^"
            r"(?P<perms>[-bcdlps][-rwxsStT]{9})\s+"
            r"(?P<links>\d+)\s"
            r"(?P<uid>\d+)\s+"
            r"(?P<gid>\d+)\s+"
            r"(?P<size>\d+)\s"
            r"(?P<date>\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}\s\+\d{4})\s"
            r"(?P<name>.*)",
        },
        "toolbox": {
            "ls": "toolbox ls -anl {}",
            "rls": "toolbox ls -Ranl {}",
            "date_re": "%Y-%m-%d %H:%M",
            "file_re": r"^(?P<perms>[-bcdlps][-rwxsStT]{9})\s+"
            r"(?P<links>\d+)\s"
            r"(?P<uid>\d+)\s+"
            r"(?P<gid>\d+)\s+"
            r"(?P<size>\d+)?\s"
            r"(?P<date>\d{4}-\d{2}-\d{2}\s\d{2}:\d{2})\s"
            r"(?P<name>.*)",
        },
        "toybox": {
            "ls": "toybox ls -anLc {}",
            "rls": "toybox ls -Ranl {}",
            "date_re": "%Y-%m-%d %H:%M",
            "file_re": r"^(?P<perms>[-bcdlps][-rwxsStT]{9})\s+"
            r"(?P<links>\d+)\s+"
            r"(?P<uid>\d+)\s+"
            r"(?P<gid>\d+)\s+"
            r"(?P<size>\d+)?\s"
            r"(?P<date>\d{4}-\d{2}-\d{2}\s\d{2}:\d{2})\s"
            r"(?P<name>.*)",
        },
    }

    def __init__(self, config, dev):
        box = config.box
        if box == "default":
            for name in (
                "toybox",
                "busybox",
                "toolbox",
            ):
                line = dev.shell("which " + name).strip()
                if line:
                    box = name
                    break
            else:
                raise IOError(2, "unknown shell")
        self.box = self.boxes[box]
        self.file_re = re.compile(self.box["file_re"])

    def list(self, dev, top):
        dtop = top
        if dtop[-1] != "/":
            dtop += "/"
        cmd = self.box["ls"].format(dtop)
        lines = dev.shell(cmd)
        result = []
        for line in lines.split("\n"):
            line = line.strip()
            if not line:
                continue
            rm = self.file_re.match(line)
            if not rm:
                if line.split()[0] != "total":
                    log.debug("ignored entry in ({}): {}".format(top, line))
                continue
            entry = Entry(top, **rm.groupdict())
            entry.update(self.box)
            if entry.perms[0] == "l":
                self.listfix(dev, entry, entry.link_target)
            result.append(entry)
        return result

    def listfix(self, dev, entry, target):
        lines = dev.shell("stat " + target).split("\n")
        target = lines[0].split(" -> ")
        if len(target) == 2:
            target = target[1].strip()[1:-1]
            return self.listfix(dev, entry, target)
        try:
            perms = lines[3].split(")")[0]
            perms = perms.split("(")[1]
            perms = perms.split("/")[1]
            if perms[0] != "d":
                entry.size = int(lines[1].split()[1])
            entry.perms = perms
            entry.perms2mode()
        except IndexError:
            # link to non existing element
            pass


class Plugin(PluginVFS):
    label = "Python ADB"
    openFrom = ["PLUGINSMENU", "DISKMENU"]

    def __init__(self, parent, info, ffi, ffic):
        super().__init__(parent, info, ffi, ffic)
        self.config = Config(self.USERHOME)

    def OpenPlugin(self, OpenFrom):
        self.names = []
        self.Items = []
        self.clt = AdbClient()
        self.deviceserial = None
        self.device = None
        self.devicepath = "/"
        return True

    def loadDevices(self):
        for d in self.clt.list():
            self.addName(d.serial, FILE_ATTRIBUTE_DIRECTORY, 0)

    def setName(self, i, name, attr, size):
        self.names[i] = self.s2f(name)
        item = self.Items[i].FindData
        item.dwFileAttributes = attr
        item.nFileSize = size
        item.lpwszFileName = self.names[i]
        return item

    def addName(self, name, attr, size):
        # log.debug('addName({})'.format(name))
        # increase C array
        items = self.ffi.new("struct PluginPanelItem []", len(self.Items) + 1)
        for i in range(len(self.Items)):
            items[i] = self.Items[i]
        self.Items = items
        self.names.append(None)
        self.setName(len(self.Items) - 1, name, attr, size)

    def addResult(self, result):
        n = 0
        for rec in result:
            if rec.name in (".", ".."):
                continue
            n += 1
        self.Items = self.ffi.new("struct PluginPanelItem []", n)
        self.names = [None] * n
        i = 0
        for rec in result:
            if rec.name in (".", ".."):
                continue
            name = rec.name.replace('\\ ', ' ')
            attr = 0
            attr |= FILE_ATTRIBUTE_DIRECTORY if rec.mode & stat.S_IFDIR else 0
            attr |= FILE_ATTRIBUTE_DEVICE if rec.mode & stat.S_IFCHR else 0
            attr |= FILE_ATTRIBUTE_ARCHIVE if rec.mode & stat.S_IFREG else 0
            # log.debug('{} mode={:5o} attr={} perms={}'.format(name, rec.mode, attr, rec.perms))
            # datetime.datetime.fromtimestamp(rec.time).strftime('%Y-%m-%d %H:%M:%S')
            item = self.setName(i, name, attr, rec.size)
            item.dwUnixMode = rec.mode

            t = (
                int(time.mktime(rec.date.timetuple())) + EPOCH_DIFFERENCE
            ) * TICKS_PER_SECOND
            item.ftLastWriteTime.dwHighDateTime = t >> 32
            item.ftLastWriteTime.dwLowDateTime = t & 0xFFFFFFFF
            i += 1

    def deleteNames(self, names):
        found = []
        for i in range(len(self.names)):
            if self.f2s(self.names[i]) in names:
                found.append(i)
        if len(found) == 0:
            return False
        # new array
        n = len(self.Items) - len(found)
        if n > 0:
            items = self.ffi.new("struct PluginPanelItem []", n)
            j = 0
            for i in range(len(self.Items)):
                if i not in found:
                    items[j] = self.Items[i]
                    j += 1
            # delete
            for i in sorted(found, reverse=True):
                del self.names[i]
        else:
            self.names = []
            items = []
        self.Items = items
        self.info.Control(self.hplugin, self.ffic.FCTL_UPDATEPANEL, 0, 0)
        self.info.Control(self.hplugin, self.ffic.FCTL_REDRAWPANEL, 0, 0)
        return True

    def Message(self, lines):
        _MsgItems = [
            self.s2f("ADB"),
            self.s2f(""),
        ]
        for line in lines:
            _MsgItems.append(self.s2f(line))
        _MsgItems.extend(
            [
                self.s2f(""),
                self.s2f("\x01"),
                self.s2f("&Ok"),
            ]
        )
        # log.debug('_msgItems: %s', _MsgItems)
        MsgItems = self.ffi.new("wchar_t *[]", _MsgItems)
        self.info.Message(
            self.info.ModuleNumber,  # GUID
            self.ffic.FMSG_WARNING | self.ffic.FMSG_LEFTALIGN,  # Flags
            self.s2f("Contents"),  # HelpTopic
            MsgItems,  # Items
            len(MsgItems),  # ItemsNumber
            1,  # ButtonsNumber
        )

    def ConfirmDialog(self, title, lines):
        _MsgItems = [
            self.s2f(title),
            self.s2f(""),
        ]
        for line in lines:
            _MsgItems.append(self.s2f(line))
        _MsgItems.extend(
            [
                self.s2f(""),
                self.s2f("\x01"),
                self.s2f("&Cancel"),
                self.s2f("&Ok"),
            ]
        )
        # log.debug('_msgItems: %s', _MsgItems)
        MsgItems = self.ffi.new("wchar_t *[]", _MsgItems)
        rc = self.info.Message(
            self.info.ModuleNumber,  # GUID
            self.ffic.FMSG_WARNING | self.ffic.FMSG_LEFTALIGN,  # Flags
            self.s2f("Contents"),  # HelpTopic
            MsgItems,  # Items
            len(MsgItems),  # ItemsNumber
            2,  # ButtonsNumber
        )
        # 0 = Cancel, 1 = OK
        # log.debug('confirm={}'.format(rc))
        return rc

    @staticmethod
    def HandleCommandLine(line):
        return line in ("uadb",)

    def CommandLine(self, line):
        sline = line.split()
        if len(sline) == 2 and sline[1] == "config":
            self.Configure()
        else:
            self.Message(["Unknown command line:", line])

    def Configure(self):
        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                try:
                    dlgid = {
                        "default": dlg.ID_default,
                        "busybox": dlg.ID_busybox,
                        "toybox": dlg.ID_toybox,
                        "toolbox": dlg.ID_toolbox,
                    }[self.config.box]
                    dlg.SetCheck(dlgid, self.ffic.BSTATE_CHECKED)
                    dlg.SetCheck(
                        dlg.ID_rexec,
                        self.ffic.BSTATE_CHECKED if self.config.rexec else 0,
                    )
                except:
                    log.exception("bang")
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = DialogBuilder(
            self,
            DialogProc,
            "ADB Config",
            "adb config",
            0,
            VSizer(
                HSizer(
                    VSizer(
                        TEXT(None, "Shell:"),
                        RADIOBUTTON("default", "default", flags=self.ffic.DIF_GROUP),
                        RADIOBUTTON("busybox", "busybox"),
                        RADIOBUTTON("toybox", "toybox"),
                        RADIOBUTTON("toolbox", "toolbox"),
                    ),
                    HSizer(
                        CHECKBOX("rexec", "Allow remote execute"),
                    ),
                ),
                HLine(),
                HSizer(
                    BUTTON("OK", "OK", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("CANCEL", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == dlg.ID_OK:
            box = 0
            for n, did in (
                (0, dlg.ID_default),
                (1, dlg.ID_busybox),
                (2, dlg.ID_toybox),
                (3, dlg.ID_toolbox),
            ):
                box |= n if dlg.GetCheck(did) else 0
            self.config.box = {
                0: "default",
                1: "busybox",
                2: "toybox",
                3: "toolbox",
            }[box]
            self.config.rexec = True if dlg.GetCheck(dlg.ID_rexec) else False
            self.config.save()

    def GetOpenPluginInfo(self, OpenInfo):
        if self.device is not None:
            title = "adb://{}{}".format(self.device.serial, self.devicepath)
        else:
            title = self.label
        #
        # WARNING WARNING - dangerous pointers and lifetime of a variable in python
        # anything passed as a pointer must be kept in local python variables
        # otherwise gc/cffi will delete them and miracles will happen
        #
        self._curdir = self.s2f(self.devicepath)
        self._title = self.s2f(title)
        self._label = self.s2f(self.label)

        Info = self.ffi.cast("struct OpenPluginInfo *", OpenInfo)
        Info.Flags = (
            self.ffic.OPIF_USEFILTER
            | self.ffic.OPIF_USESORTGROUPS
            | self.ffic.OPIF_USEHIGHLIGHTING
            | self.ffic.OPIF_ADDDOTS
            | self.ffic.OPIF_SHOWNAMESONLY
        )
        Info.HostFile = self.ffi.NULL
        Info.CurDir = self._curdir
        Info.Format = self._label
        Info.PanelTitle = self._title
        # const struct InfoPanelLine *InfoLines;
        # int                   InfoLinesNumber;
        # const wchar_t * const   *DescrFiles;
        # int                   DescrFilesNumber;
        # const struct PanelMode *PanelModesArray;
        Info.PanelModesNumber = 0
        Info.StartPanelMode = 0
        # Info.StartSortMode = self.ffic.SM_NAME
        # Info.StartSortOrder = 0
        # const struct KeyBarTitles *KeyBar;
        # Info.ShortcutData = self.s2f('py:adb cd '+title)

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        # super().GetFindData(PanelItem, ItemsNumber, OpMode)
        if OpMode & self.ffic.OPM_FIND:
            # find in root of device causes core dump on linux
            return False
        if self.device is None:
            try:
                self.loadDevices()
            except RuntimeError as why:
                log.error(why)
                lines = str(why).split("\n")
                lines.insert(0, "RuntimeError:")
                self.Message(lines)
                return False
            except AdbError as ex:
                log.exception("AdbError")
                self.Message(str(ex).split("\n"))
                return False
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        else:
            try:
                shell = Shell(self.config, self.device)
                result = shell.list(self.device, self.devicepath)
                self.addResult(result)
            except AdbError as ex:
                log.exception("AdbError")
                self.Message(str(ex).split("\n"))
                return False
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        PanelItem = self.ffi.cast("struct PluginPanelItem **", PanelItem)
        ItemsNumber = self.ffi.cast("int *", ItemsNumber)
        n = len(self.Items)
        p = self.ffi.NULL if n == 0 else self.Items
        PanelItem[0] = p
        ItemsNumber[0] = n
        # log.debug("ADB.GetFindData, p={} n={}".format(p, n))
        return True

    def FreeFindData(self, PanelItem, ItemsNumber):
        # log.debug("ADB.FreeFindData({0}, {1}, n.names={2}, n.Items={3})".format(PanelItem, ItemsNumber, len(self.names), len(self.Items)))
        self.names = []
        self.Items = []

    def SetDirectory(self, Dir, OpMode):
        # super().SetDirectory(Dir, OpMode)
        # if OpMode & self.ffic.OPM_FIND:
        #    return False
        name = self.f2s(Dir)
        # log.debug('goto.0: devicepath={} name={}'.format(self.devicepath, name))
        if name == "":
            self.info.Control(self.hplugin, self.ffic.FCTL_CLOSEPLUGIN, 0, 0)
            return True
        # log.debug('goto.1: devicepath={} name={}'.format(self.devicepath, name))
        if name == "..":
            if self.device is None:
                self.info.Control(self.hplugin, self.ffic.FCTL_CLOSEPLUGIN, 0, Dir)
                return True
            if self.devicepath == "/":
                self.device = None
            else:
                self.devicepath = os.path.normpath(os.path.join(self.devicepath, name))
            # log.debug('goto.2: devicepath={}'.format(self.devicepath))
        elif self.device is None:
            try:
                self.device = self.clt.device(name)
                self.devicepath = "/"
                try:
                    self.device.root()
                    # log.debug('goto.3: rooted')
                except:
                    # log.debug('goto.3: not rooted')
                    pass
            except AdbError as ex:
                self.device = None
                log.exception("AdbError")
                self.Message(str(ex).split("\n"))
                return False
            except Exception as ex:
                self.device = None
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        else:
            self.devicepath = os.path.join(self.devicepath, name)
        return True

    def Compare(self, PanelItem1, PanelItem2, Mode):
        def cmp(a, b):
            if a < b:
                return -1
            elif a == b:
                return 0
            return 1

        p1 = self.ffi.cast("struct PluginPanelItem *", PanelItem1)
        p2 = self.ffi.cast("struct PluginPanelItem *", PanelItem2)
        n1 = self.f2s(p1.FindData.lpwszFileName)
        n2 = self.f2s(p2.FindData.lpwszFileName)
        rc = cmp(n1, n2)
        # log.debug('Compare: cmp({}, {})={}, mode={}'.format(n1, n2, rc, Mode))
        return rc

    def GetVirtualFindData(self, PanelItem, ItemsNumber, Path):
        return False

    def FreeVirtualFindData(self, PanelItem, ItemsNumber):
        log.debug("FreeVirtualFindData: {}, {}".format(PanelItem, ItemsNumber))

    def ProcessEvent(self, Event, Param):
        # if Event == self.ffic.FE_IDLE:
        pass

    def GetFiles(self, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        # super().GetFiles(PanelItem, ItemsNumber, Move, DestPath, OpMode)
        if ItemsNumber == 0 or Move:
            return False
        if self.device is None:
            self.Message(["GetFiles allowed only inside device."])
            return True
        # TODO dialog with copy parameters
        # TODO progress dialog
        # TODO copy in background
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        DestPath = self.ffi.cast("wchar_t **", DestPath)
        dpath = self.ffi.string(DestPath[0])
        # log.debug('GetFiles: {} {} OpMode={}'.format(ItemsNumber, OpMode, dpath))
        for i in range(ItemsNumber):
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            sqname = os.path.join(self.devicepath, name)
            dqname = os.path.join(dpath, name)
            # log.debug('pull: {} -> {} OpMode={}'.format(sqname, dqname, OpMode))
            try:
                self.device.sync.pull(sqname, dqname)
            except AdbError as ex:
                log.exception("AdbError")
                self.Message(str(ex).split("\n"))
                return False
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        return True

    def PutFiles(self, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        # super().PutFiles(PanelItem, ItemsNumber, Move, SrcPath, OpMode)
        if ItemsNumber == 0 or Move:
            return False
        if self.device is None:
            self.Message(["PutFiles allowed only inside device."])
            return True
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        spath = self.f2s(SrcPath)
        for i in range(ItemsNumber):
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            sqname = os.path.join(spath, name)
            dqname = os.path.join(self.devicepath, name)
            # log.debug('push: {} -> {} OpMode={}'.format(sqname, dqname, OpMode))
            try:
                self.device.sync.push(sqname, dqname)
            except AdbError as ex:
                log.exception("AdbError")
                self.Message(str(ex).split("\n"))
                return False
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        return True

    def DeleteFiles(self, PanelItem, ItemsNumber, OpMode):
        # super().DeleteFiles(PanelItem, ItemsNumber, OpMode)
        if ItemsNumber == 0:
            return False
        if self.device is None:
            self.Message(["DeleteFiles allowed only inside device."])
            return True
        items = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        if ItemsNumber > 1:
            rc = self.ConfirmDialog("Delete", ["Do you wish to delete the files"])
        else:
            name = self.f2s(items[0].FindData.lpwszFileName)
            rc = self.ConfirmDialog("Delete", ["Do you wish to delete the file", name])
        if rc != 1:
            return True
        for i in range(ItemsNumber):
            name = self.f2s(items[i].FindData.lpwszFileName)
            if name in (".", ".."):
                continue
            dqname = os.path.join(self.devicepath, name)
            log.debug("remove: {}, OpMode={}".format(dqname, OpMode))
            try:
                st = self.device.sync.stat(dqname)
                if stat.S_ISDIR(st.mode):
                    s = self.device.shell("rmdir " + dqname)
                else:
                    s = self.device.shell("rm " + dqname)
                if s:
                    log.debug("result:" + s)
                    self.Message(s.split("\n"))
            except AdbError as ex:
                log.exception("AdbError")
                self.Message(str(ex).split("\n"))
                return False
            except Exception as ex:
                log.exception("unknown exception:")
                msg = str(ex).split("\n")
                msg.insert(0, "Unknown exception.")
                self.Message(msg)
                return False
        return True

    def MakeDirectory(self, Name, OpMode):
        # super().DMakeDirectory(Name, OpMode)
        if self.device is None:
            self.Message(["Make directory allowed only inside device."])
            return True
        name = self.ffi.cast("wchar_t **", Name)
        name = self.ffi.string(name[0])

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                log.debug("INITDIALOG")
                try:
                    dlg.SetText(dlg.ID_path, name)
                    dlg.SetFocus(dlg.ID_path)
                except:
                    log.exception("bang")
                log.debug("/INITDIALOG")
            elif Msg == self.ffic.DN_BTNCLICK:
                pass
            elif Msg == self.ffic.DN_KEY:
                if Param2 == self.ffic.KEY_ESC:
                    return 0
                if Param2 == self.ffic.KEY_LEFT:
                    pass
                elif Param2 == self.ffic.KEY_UP:
                    pass
                elif Param2 == self.ffic.KEY_RIGHT:
                    pass
                elif Param2 == self.ffic.KEY_DOWN:
                    pass
                elif Param2 == self.ffic.KEY_ENTER:
                    pass
                elif Param2 == self.ffic.KEY_ESC:
                    pass
            elif Msg == self.ffic.DN_MOUSECLICK:
                pass
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = DialogBuilder(
            self,
            DialogProc,
            "Make folder",
            "makefolder",
            0,
            VSizer(
                TEXT(None, "Create the &folder"),
                EDIT("path", 36, maxlength=40),
                HLine(),
                HSizer(
                    BUTTON("OK", "OK", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("CANCEL", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == dlg.ID_OK:
            path = dlg.GetText(dlg.ID_path).strip()
            path = os.path.normpath(os.path.join(self.devicepath, path))
            log.debug("mkdir {}".format(path))
            if path:
                try:
                    s = self.device.shell("mkdir " + path)
                    if s:
                        log.debug("result:" + s)
                        self.Message(s.split("\n"))
                except AdbError as ex:
                    log.exception("AdbError")
                    self.Message(str(ex).split("\n"))
                    return False
                except Exception as ex:
                    log.exception("unknown exception:")
                    msg = str(ex).split("\n")
                    msg.insert(0, "Unknown exception.")
                    self.Message(msg)
                    return False
        self.info.DialogFree(dlg.hDlg)

        return True

    def Execute(self):
        pass

    def ProcessKey(self, Key, ControlState):
        # log.debug("ProcessKey({0}, {1})".format(Key, ControlState))
        if ControlState == self.ffic.PKF_CONTROL:
            log.debug("ProcessKey(0x{:x}, {})".format(Key, ControlState))
        # if (
        #    Key == 0x80051 # FARMACRO_KEY_EVENT = 0x80000|1...
        #    and ControlState == self.ffic.PKF_CONTROL
        # ):
        #    # 0x80051 = CTRL+Q = quick view, here its equal to core dump when scanning /proc directory
        #    return True
        if (
            False
            and Key == self.ffic.KEY_CTRLA - self.ffic.KEY_CTRL
            and ControlState == self.ffic.PKF_CONTROL
        ):
            log.debug("ProcessKey: CTRL+A")
            # self.EditAttributes()
            return True
        if (
            Key == self.ffic.KEY_ENTER
            and ControlState == self.ffic.PKF_CONTROL
            and self.config.rexec
        ):
            log.debug("ProcessKey: CTRL+ENTER")
            self.Execute()
            return True
        return False



import os
import logging
from far2l.plugin import PluginVFS


log = logging.getLogger(__name__)


class Screen:
    def __init__(self, parent):
        self.parent = parent
        self.hScreen = None

    def __enter__(self):
        self.hScreen = self.parent.info.SaveScreen(0, 0, -1, -1)

    def __exit__(self, *args):
        self.parent.info.RestoreScreen(self.hScreen)


class Plugin(PluginVFS):
    label = "Python upanel"
    openFrom = ["PLUGINSMENU", "DISKMENU"]

    def OpenPlugin(self, OpenFrom):
        self.Root = "PythonPanel"
        self.names = []
        self.Items = []
        return True

    def GetOpenPluginInfo(self, OpenInfo):
        # dangeling refs
        self._Root = self.s2f(self.Root)
        self._label = self.s2f(self.label)
        # /dangeling refs
        Info = self.ffi.cast("struct OpenPluginInfo *", OpenInfo)
        Info.Flags = (
            self.ffic.OPIF_USEFILTER
            | self.ffic.OPIF_USESORTGROUPS
            | self.ffic.OPIF_USEHIGHLIGHTING
            | self.ffic.OPIF_ADDDOTS
            | self.ffic.OPIF_SHOWNAMESONLY
        )
        Info.HostFile = self.ffi.NULL
        Info.CurDir = self._Root
        Info.Format = self._label
        Info.PanelTitle = self._label

    def GetFindData(self, PanelItem, ItemsNumber, OpMode):
        log.debug(
            "VFS.GetFindData({0}, {1}, {2})".format(PanelItem, ItemsNumber, OpMode)
        )
        PanelItem = self.ffi.cast("struct PluginPanelItem **", PanelItem)
        ItemsNumber = self.ffi.cast("int *", ItemsNumber)
        n = len(self.Items)
        p = self.ffi.NULL if n == 0 else self.Items
        PanelItem[0] = p
        ItemsNumber[0] = n
        return 1

    def SetDirectory(self, Dir, OpMode):
        if OpMode & self.ffic.OPM_FIND:
            return 0
        self.info.Control(self.hplugin, self.ffic.FCTL_CLOSEPLUGIN, 0, 0)
        return 1

    def PutFiles(self, PanelItem, ItemsNumber, Move, SrcPath, OpMode):
        PanelItem = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        SrcPath = self.f2s(SrcPath)
        for i in range(ItemsNumber):
            self.PutFile(PanelItem[i], Move, SrcPath, OpMode)
        return 1

    def PutFile(self, PanelItem, Move, SrcPath, OpMode):
        fqname = os.path.join(SrcPath, self.f2s(PanelItem.FindData.lpwszFileName))
        # increase C array
        items = self.ffi.new("struct PluginPanelItem []", len(self.Items) + 1)
        for i in range(len(self.Items)):
            items[i] = self.Items[i]
        self.Items = items
        # append element
        self.names.append(self.s2f(fqname))
        i = len(self.Items) - 1
        items[i].FindData = PanelItem.FindData
        items[i].FindData.lpwszFileName = self.names[i]

    def DeleteFiles(self, PanelItem, ItemsNumber, OpMode):
        item = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        snames = []
        for i in range(ItemsNumber):
            snames.append(self.f2s(item[i].FindData.lpwszFileName))
        found = []
        for i in range(len(self.names)):
            if self.f2s(self.names[i]) in snames:
                found.append(i)
        if len(found) == 0:
            return 0
        # new array
        items = self.ffi.new("struct PluginPanelItem []", len(self.Items) - len(found))
        j = 0
        for i in range(len(self.Items)):
            if i not in found:
                items[j] = self.Items[i]
                j += 1
        # delete
        for i in sorted(found, reverse=True):
            del self.names[i]
        self.Items = items
        self.info.Control(self.hplugin, self.ffic.FCTL_UPDATEPANEL, 0, 0)
        self.info.Control(self.hplugin, self.ffic.FCTL_REDRAWPANEL, 0, 0)
        return 0

    def GetFiles(self, PanelItem, ItemsNumber, Move, DestPath, OpMode):
        if ItemsNumber == 0 or Move:
            return 0
        item = self.ffi.cast("struct PluginPanelItem *", PanelItem)
        DestPath = self.ffi.cast("wchar_t **", DestPath)
        dpath = self.ffi.string(DestPath[0])
        for i in range(ItemsNumber):
            sqname = self.f2s(item[i].FindData.lpwszFileName)
            dqname = os.path.join(dpath, sqname.split("/")[-1])
            # log.debug('GetFiles OpMode={} source={} destination={}'.format(OpMode, sqname, dqname))
            # just copy file, local filesystem only
            data = open(sqname, "rb").read()
            open(dqname, "wb").write(data)
        return 1



import logging
import time
import re
from far2l.plugin import PluginBase
from far2l.farprogress import ProgressMessage
from far2l.fardialogbuilder import (
    TEXT,
    EDIT,
    BUTTON,
    CHECKBOX,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)


log = logging.getLogger(__name__)


class Worker:
    def __init__(self, parent, curline, curpos, nlines, pat, repl):
        self.parent = parent
        self.curline = curline
        self.curpos = curpos
        self.nlines = nlines
        self.pat = pat
        self.repl = repl
        self.prog = re.compile(pat)

    def info(self, line):
        print('***** curline=', self.curline, 'curpos=', self.curpos, '**', line)

    def replace(self, spart, m):
        repl = self.prog.sub(self.repl, spart[m.start():m.end()], count=1)
        rc = self.parent.ask(m, spart[m.start():m.end()], repl)
        if rc == 0:
            spart = spart[:m.start()] + repl + spart[m.end():]
        return rc, spart

    def forward(self):
        while self.curline < self.nlines:
            line = self.parent.getline(self.curline)
            self.info(line)
            while True:
                spart = line[self.curpos:]
                m = self.prog.search(spart)
                if not m:
                    break
                rc, repl = self.replace(spart, m)
                if rc == 3:
                    return
                line = line[:self.curpos] + repl
                self.curpos += len(repl)-(len(spart)-m.end())
                self.parent.setline(self.curline, line)
            self.curline += 1
            self.curpos = 0

    def backward(self):
        while self.curline >= 0:
            line = self.parent.getline(self.curline)
            self.info(line)
            while True:
                spart = line[:self.curpos]
                ml = list(self.prog.finditer(spart))
                if not ml:
                    break
                rc, repl = self.replace(spart, ml[-1])
                if rc == 3:
                    return
                line = repl+line[self.curpos:]
                self.curpos = ml[-1].start()
                self.parent.setline(self.curline, line)
            self.curline -= 1
            self.curpos = -1

class Plugin(PluginBase):
    label = "Python ED Replace"
    openFrom = []
    f_search = ''
    f_replace = ''
    f_case = 0
    f_whole = 0
    f_rev = 0
    f_rx = 0
    f_select = 0
    f_is_search = False

    def ProcessEditorInput(self, Rec):
        rec = self.ffi.cast("INPUT_RECORD *", Rec)

        if (
            rec.EventType != self.ffic.KEY_EVENT
            or not rec.Event.KeyEvent.bKeyDown
            or (rec.Event.KeyEvent.wVirtualKeyCode != self.ffic.VK_F3)
        ):
            return 0
        ks = rec.Event.KeyEvent.dwControlKeyState
        fctrl = ks & (self.ffic.RIGHT_CTRL_PRESSED | self.ffic.LEFT_CTRL_PRESSED) != 0
        falt = ks & (self.ffic.RIGHT_ALT_PRESSED | self.ffic.LEFT_ALT_PRESSED) != 0
        fshift = ks & (self.ffic.SHIFT_PRESSED) != 0
        log.debug(f'ED Replace ctrl:{fctrl} alt:{falt} shift:{fshift}')
        self.Perform(fctrl, falt, fshift)
        return 1

    def ConfigDialog(self, fctrl):
        self.f_is_search = not fctrl

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                try:
                    dlg.SetCheck(dlg.ID_f_case, self.f_case)
                    dlg.SetCheck(dlg.ID_f_whole, self.f_whole)
                    dlg.SetCheck(dlg.ID_f_rev, self.f_rev)
                    dlg.SetCheck(dlg.ID_f_rx, self.f_rx)
                    dlg.SetCheck(dlg.ID_f_select, self.f_select)
                except:
                    log.exception("bang")
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        title = "Search" if self.f_is_search else "Replace"
        dlgitems = [
            HSizer(
                TEXT(None, "Search &for  :"),
                EDIT("f_search", 33, maxlength=80, history='f_search', flags=self.ffic.DIF_USELASTHISTORY|self.ffic.DIF_HISTORY),
            )
        ]
        if not self.f_is_search:
            dlgitems.append(
                HSizer(
                    TEXT(None, "Replace &with:"),
                    EDIT("f_replace", 33, maxlength=80, history='f_replace', flags=self.ffic.DIF_USELASTHISTORY|self.ffic.DIF_HISTORY),
                )
            )
        dlgitems.extend([
            HLine(),
            HSizer(
                VSizer(
                    CHECKBOX("f_case", "&Case sensitive"),
                    CHECKBOX("f_whole", "W&hole words"),
                    CHECKBOX("f_rev", "Re&verse search"),
                ),
                VSizer(
                    CHECKBOX("f_rx", "Regular e&xpressions"),
                    CHECKBOX("f_select", "Se&lect found"),
                )
            ),
            HLine(),
            HSizer(
                BUTTON("vreplace", "&"+title, default=True, flags=self.ffic.DIF_CENTERGROUP),
                BUTTON("vcancel", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
            ),
        ])
        b = DialogBuilder(
            self,
            DialogProc,
            title,
            "uedreplace",
            0,
            VSizer(*list(dlgitems)),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == dlg.ID_vreplace:
            self.f_search = dlg.GetText(dlg.ID_f_search)
            if self.f_is_search:
                self.f_replace = ''
            else:
                self.f_replace = dlg.GetText(dlg.ID_f_replace)
            self.f_case = dlg.GetCheck(dlg.ID_f_case)
            self.f_whole = dlg.GetCheck(dlg.ID_f_whole)
            self.f_rev = dlg.GetCheck(dlg.ID_f_rev)
            self.f_rx = dlg.GetCheck(dlg.ID_f_rx)
            self.f_select = dlg.GetCheck(dlg.ID_f_select)
        self.info.DialogFree(dlg.hDlg)
        log.debug(f'ED Replace cfg: res={res} rep={dlg.ID_vreplace}')
        return res == dlg.ID_vreplace

    def Ask(self, sold, snew):
        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)
        b = DialogBuilder(
            self,
            DialogProc,
            "Replace",
            "uedreplace",
            0,
            VSizer(
                TEXT(None, "Replace:"+sold.strip()),
                TEXT(None, "With   :"+snew.strip()),
                HLine(),
                HSizer(
                    BUTTON("vreplace", "&Replace", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vall", "&All", flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vskip", "&Skip", flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vcancel", "&Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        self.info.DialogFree(dlg.hDlg)
        if res == dlg.ID_vreplace:
            return 0
        elif res == dlg.ID_vall:
            return 1
        elif res == dlg.ID_vskip:
            return 2
        elif res == dlg.ID_vcancel:
            return 3
        return 3

    def Perform(self, fctrl, falt, fshift):
        if not self.f_search or not (falt or fshift):
            if not self.ConfigDialog(fctrl):
                return
            if not self.f_search:
                return
        if falt:
            self.DoPrev()
        elif fshift:
            self.DoNext()
        else:
            if self.f_rev:
                self.DoPrev()
            else:
                self.DoNext()

    def DoNext(self):
        f_search = self.f_search
        f_replace = self.f_replace
        if not self.f_rx and self.f_whole:
            f_search = r'\b'+f_search+r'\b'
        f_flags = re.IGNORECASE if not self.f_case else 0
        prog = re.compile(f_search, flags=f_flags)

        ei = self.ffi.new("struct EditorInfo *")
        self.info.EditorControl(self.ffic.ECTL_GETINFO, ei)

        self.fask = True
        self.found = False

        egs = self.ffi.new("struct EditorGetString *")
        esp = self.ffi.new("struct EditorSetPosition *")
        esp.CurTabPos = -1
        esp.TopScreenLine = -1
        esp.LeftPos = -1
        esp.Overtype = -1
        es = self.ffi.new("struct EditorSelect *")

        class IO:
            lno = None
            line = None

            def getline(me, lno):
                if lno == me.lno:
                    return me.line
                egs.StringNumber = lno
                self.info.EditorControl(self.ffic.ECTL_GETSTRING, egs)
                me.line = self.f2s(egs.StringText)
                return me.line

            def setline(me, lno, line):
                # select whole line
                es.BlockType = self.ffic.BTYPE_COLUMN
                es.BlockStartLine = lno
                es.BlockStartPos = 0
                es.BlockWidth = len(me.line)
                es.BlockHeight = 1
                self.info.EditorControl(self.ffic.ECTL_SELECT, es)
                # delete selected line
                self.info.EditorControl(self.ffic.ECTL_DELETEBLOCK, self.ffi.NULL)
                # insert new line
                self.info.EditorControl(self.ffic.ECTL_INSERTTEXT, self.s2f(line))
                me.line = line

            def ask(me, m, found, replace):
                self.found = True
                if self.f_is_search or self.Ask:
                    # set editor position
                    esp.CurLine = worker.curline
                    esp.CurPos = worker.curpos+m.start()
                    self.info.EditorControl(self.ffic.ECTL_SETPOSITION, esp)
                    if self.f_select or self.Ask:
                        # select found item
                        es.BlockType = self.ffic.BTYPE_COLUMN
                        es.BlockStartLine = worker.curline
                        es.BlockStartPos = worker.curpos+m.start()
                        es.BlockWidth = m.end()-m.start()
                        es.BlockHeight = 1
                        self.info.EditorControl(self.ffic.ECTL_SELECT, es)
                    # refresh
                    self.info.EditorControl(self.ffic.ECTL_REDRAW, self.ffi.NULL)
                if self.f_is_search:
                    # cancel=search only
                    return 3
                if self.fask:
                    # ask for decision
                    rc = self.Ask(found, replace)
                else:
                    rc = 0
                if rc == 1:
                    # all
                    self.fask = False
                    rc = 0
                return rc

        if not self.f_is_search:
            eur = self.ffi.new("struct EditorUndoRedo *")
            eur.Command = self.ffic.EUR_BEGIN
            self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        worker = Worker(IO(), ei.CurLine, ei.CurPos, ei.TotalLines, prog, f_replace)
        worker.forward()

        if not self.f_is_search:
            eur = self.ffi.new("struct EditorUndoRedo *")
            eur.Command = self.ffic.EUR_END
            self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        if not self.found:
            self.notice("Could not find the string:"+f_search)

        self.info.EditorControl(self.ffic.ECTL_REDRAW, self.ffi.NULL)

    def DoPrev(self):
        f_search = self.f_search
        f_replace = self.f_replace
        if not self.f_rx and self.f_whole:
            f_search = r'\b'+f_search+r'\b'
        f_flags = re.IGNORECASE if not self.f_case else 0
        prog = re.compile(f_search, flags=f_flags)

        ei = self.ffi.new("struct EditorInfo *")
        self.info.EditorControl(self.ffic.ECTL_GETINFO, ei)

        self.fask = True
        self.found = False

        egs = self.ffi.new("struct EditorGetString *")
        esp = self.ffi.new("struct EditorSetPosition *")
        esp.CurTabPos = -1
        esp.TopScreenLine = -1
        esp.LeftPos = -1
        esp.Overtype = -1
        es = self.ffi.new("struct EditorSelect *")

        class IO:
            lno = None
            line = None

            def getline(me, lno):
                if lno == me.lno:
                    return me.line
                egs.StringNumber = lno
                self.info.EditorControl(self.ffic.ECTL_GETSTRING, egs)
                me.line = self.f2s(egs.StringText)
                return me.line

            def setline(me, lno, line):
                # select whole line
                es.BlockType = self.ffic.BTYPE_COLUMN
                es.BlockStartLine = lno
                es.BlockStartPos = 0
                es.BlockWidth = len(me.line)
                es.BlockHeight = 1
                self.info.EditorControl(self.ffic.ECTL_SELECT, es)
                # delete selected line
                self.info.EditorControl(self.ffic.ECTL_DELETEBLOCK, self.ffi.NULL)
                # insert new line
                self.info.EditorControl(self.ffic.ECTL_INSERTTEXT, self.s2f(line))
                me.line = line

            def ask(me, m, found, replace):
                self.found = True
                if self.f_is_search or self.Ask:
                    # set editor position
                    esp.CurLine = worker.curline
                    esp.CurPos = m.start()
                    self.info.EditorControl(self.ffic.ECTL_SETPOSITION, esp)
                    if self.f_select or self.Ask:
                        # select found item
                        es.BlockType = self.ffic.BTYPE_COLUMN
                        es.BlockStartLine = worker.curline
                        es.BlockStartPos = m.start()
                        es.BlockWidth = m.end()-m.start()
                        es.BlockHeight = 1
                        self.info.EditorControl(self.ffic.ECTL_SELECT, es)
                    # refresh
                    self.info.EditorControl(self.ffic.ECTL_REDRAW, self.ffi.NULL)
                if self.f_is_search:
                    # cancel=search only
                    return 3
                if self.fask:
                    # ask for decision
                    rc = self.Ask(found, replace)
                else:
                    rc = 0
                if rc == 1:
                    # all
                    self.fask = False
                    rc = 0
                return rc

        if not self.f_is_search:
            eur = self.ffi.new("struct EditorUndoRedo *")
            eur.Command = self.ffic.EUR_BEGIN
            self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        worker = Worker(IO(), ei.CurLine, ei.CurPos, ei.TotalLines, prog, f_replace)
        worker.backward()

        if not self.f_is_search:
            eur = self.ffi.new("struct EditorUndoRedo *")
            eur.Command = self.ffic.EUR_END
            self.info.EditorControl(self.ffic.ECTL_UNDOREDO, eur)

        if not self.found:
            self.notice("Could not find the string:"+f_search)

        self.info.EditorControl(self.ffic.ECTL_REDRAW, self.ffi.NULL)



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk

def show_clipboard():
    clipboard = Gdk.Display().get_default().get_clipboard()
    print("Clipboard:")
    print('str:', clipboard.get_formats().to_string())
#show_clipboard()

clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
for mime in (
    "text/plain",
    "text/plain;charset=utf-8",
    "text/html",
    "text/uri-list",
    "image/png",
    'x-special/gnome-copied-files',
    'SAVE_TARGETS',
    'TIMESTAMP',
    'TARGETS',
):
    text = clipboard.wait_for_contents(Gdk.Atom.intern(mime, True))
    if text:
        print(mime, text.get_data())



import random
import logging
from far2l.plugin import PluginBase
from far2l.fardialogbuilder import (
    TEXT,
    BUTTON,
    USERCONTROL,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)


log = logging.getLogger(__name__)


class Miner:
    def __init__(self, rows, cols, bombs):
        self.score = 0
        self.rows = rows
        self.cols = cols
        self.bombs = bombs
        self.gamemap = [[0 for col in range(self.cols)] for row in range(self.rows)]
        self.usermap = [["-" for col in range(self.cols)] for row in range(self.rows)]
        self.GenerateMap()
        self.gameover = False
        self.gamewon = False

    @property
    def done(self):
        return self.gameover or self.gamewon

    def GenerateMap(self):
        mmap = self.gamemap
        rows = self.rows
        cols = self.cols

        def mark(y, x):
            if 0 <= y < rows and 0 <= x < cols:
                if mmap[y][x] != "X":
                    mmap[y][x] += 1

        for num in range(self.bombs):
            while True:
                y = random.randint(0, rows - 1)
                x = random.randint(0, cols - 1)
                if mmap[y][x] != "X":
                    break
            mmap[y][x] = "X"
            mark(y - 1, x - 1)
            mark(y - 1, x)
            mark(y - 1, x + 1)
            mark(y, x - 1)
            mark(y, x + 1)
            mark(y + 1, x - 1)
            mark(y + 1, x)
            mark(y + 1, x + 1)

    def CheckWon(self):
        bombs = self.bombs
        for row in self.usermap:
            for cell in row:
                if cell == "-":
                    bombs -= 1
        self.gamewon = bombs == 0

    def OpenMap(self, y, x):
        ch = self.gamemap[y][x]
        if ch == "X":
            self.gameover = True
            self.gamewon = False
            return
        pmap = self.usermap
        mmap = self.gamemap
        pmap[y][x] = mmap[y][x]
        self.score += 1
        checked = {}
        stack = [(y, x)]

        def mark(y, x):
            if 0 <= y < self.rows and 0 <= x < self.cols:
                if mmap[y][x] != "X" and pmap[y][x] != mmap[y][x]:
                    self.score += 1
                    pmap[y][x] = mmap[y][x]
                stack.append((y, x))

        while len(stack):
            y, x = stack.pop()
            if checked.get((y, x)) == True:
                continue
            checked[(y, x)] = True
            if mmap[y][x] != 0:
                continue
            pmap[y][x] = 0
            mark(y - 1, x - 1)
            mark(y - 1, x)
            mark(y - 1, x + 1)
            mark(y, x - 1)
            mark(y, x + 1)
            mark(y + 1, x - 1)
            mark(y + 1, x)
            mark(y + 1, x + 1)

        self.CheckWon()


class Plugin(PluginBase):
    label = "Python uminer"
    openFrom = ["PLUGINSMENU"]

    def Redraw(self, dlg, usermap=True):
        # dlg.EnableRedraw(False)
        mapa = self.game.usermap if usermap else self.game.gamemap
        for row in range(self.game.rows):
            for col in range(self.game.cols):
                did = getattr(dlg, "ID_{}_{}".format(row, col))
                ch = str(mapa[row][col])
                dlg.SetText(did, ch)
        dlg.SetText(dlg.ID_bombs, "Bombs: {:2} ".format(self.bombs))
        msg = "Score: {:2}".format(self.game.score)
        if self.game.gameover:
            msg += ", Game Over"
        elif self.game.gamewon:
            msg += ", Game Won"
        dlg.SetText(dlg.ID_status, msg)
        # dlg.EnableRedraw(True)
        # dlg.RedrawDialog()

    def OpenPlugin(self, OpenFrom):
        self.rows = 9
        self.cols = 16
        self.bombs = 10
        self.game = Miner(self.rows, self.cols, self.bombs)
        dlg = None

        def OnButton(hDlg, Msg, Param1, Param2):
            if Param1 == dlg.ID_vshow:
                self.game.gameover = True
                self.Redraw(dlg, False)
            elif Param1 == dlg.ID_vrestart:
                self.game = Miner(self.rows, self.cols, self.bombs)
                self.Redraw(dlg)
            elif Param1 == dlg.ID_vbplus:
                self.game.gameover = True
                if self.bombs < (self.rows * self.cols) // 2:
                    self.bombs += 1
                    self.Redraw(dlg, False)
            elif Param1 == dlg.ID_vbminus:
                self.game.gameover = True
                if self.bombs > 3:
                    self.bombs -= 1
                    self.Redraw(dlg, False)
            elif Param1 == dlg.ID_vcancel:
                return self.Close(dlg.ID_vcancel)
            else:
                return False
            return True

        def OnMouseClick(hDlg, Msg, Param1, Param2):
            idmin = getattr(dlg, "ID_{}_{}".format(0, 0))
            idmax = getattr(dlg, "ID_{}_{}".format(self.rows - 1, self.cols - 1))

            if Param1 < idmin or Param1 > idmax or self.game.done:
                return False
                # return OnButton(hDlg, Msg, Param1, Param2)

            no = Param1 - 1
            row, col = divmod(no, self.game.cols)
            self.game.OpenMap(row, col)
            if self.game.gameover:
                self.game.gamemap[row][col] = "O"
                self.Redraw(dlg, False)
            else:
                self.Redraw(dlg)
            return True

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                self.Redraw(dlg)
            elif Msg == self.ffic.DN_BTNCLICK:
                return OnButton(hDlg, Msg, Param1, Param2)
            elif Msg == self.ffic.DN_MOUSECLICK:
                if not self.game.done:
                    return OnMouseClick(hDlg, Msg, Param1, Param2)
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        vs = []
        for row in range(self.game.rows):
            hs = []
            for col in range(self.game.cols):
                hs.append(TEXT("{}_{}".format(row, col), " "))
            vs.append(HSizer(*hs, border=(0, 0, 0, 0)))
        vs = VSizer(*vs)

        b = DialogBuilder(
            self,
            DialogProc,
            "Python Miner",
            "miner",
            0,
            VSizer(
                vs,
                HLine(),
                TEXT("status", " " * 25),
                HSizer(
                    BUTTON("vshow", "Show Me", flags=self.ffic.DIF_CENTERGROUP),
                    TEXT(
                        "bombs",
                        "Bombs: {:2} ".format(self.game.bombs),
                        flags=self.ffic.DIF_CENTERGROUP,
                    ),
                    BUTTON("vbplus", "+", flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vbminus", "-", flags=self.ffic.DIF_CENTERGROUP),
                ),
                HSizer(
                    BUTTON("vrestart", "Again", flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vcancel", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == -1:
            msg = "esc"
        elif res == dlg.ID_vcancel:
            msg = "cancel"
        else:
            msg = "why ?"
        log.debug("rc={} msg={}".format(res, msg))
        self.info.DialogFree(dlg.hDlg)



import logging
from far2l.plugin import PluginBase


log = logging.getLogger(__name__)


class Plugin(PluginBase):
    label = "Python udialog"
    openFrom = ["PLUGINSMENU"]

    @staticmethod
    def HandleCommandLine(line):
        return line in ("dialog", "exec")

    def CommandLine(self, line):
        if line == "dialog":
            log.debug("dialog")

            @self.ffi.callback("FARWINDOWPROC")
            def KeyDialogProc(hDlg, Msg, Param1, Param2):
                # if Msg == self.ffic.DN_KEY:
                #    if Param2 == 27:
                #        return 0
                return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

            fdis = [
                (
                    self.ffic.DI_DOUBLEBOX,
                    2,
                    1,
                    31,
                    4,
                    0,
                    {"Selected": 0},
                    0,
                    0,
                    self.s2f("Python"),
                    0,
                ),
                (
                    self.ffic.DI_TEXT,
                    -1,
                    2,
                    0,
                    2,
                    0,
                    {"Selected": 0},
                    0,
                    0,
                    self.s2f("Plugin path"),
                    0,
                ),
                (
                    self.ffic.DI_EDIT,
                    5,
                    3,
                    28,
                    3,
                    0,
                    {"Selected": 0},
                    0,
                    0,
                    self.ffi.NULL,
                    0,
                ),
            ]
            fdi = self.ffi.new("struct FarDialogItem []", fdis)
            hDlg = self.info.DialogInit(
                self.info.ModuleNumber,
                -1,
                -1,
                34,
                6,
                self.s2f("Set Python Path"),
                fdi,
                len(fdi),
                0,
                0,
                KeyDialogProc,
                0,
            )
            res = self.info.DialogRun(hDlg)
            if res != -1:
                sptr = self.info.SendDlgMessage(
                    hDlg, self.ffic.DM_GETCONSTTEXTPTR, 2, 0
                )
                path = self.f2s(sptr)
                log.debug("path: {0}".format(path))
            self.info.DialogFree(hDlg)
            log.debug("end dialog")
        else:
            line = line.split(" ", 1)[1]
            exec(line, globals(), locals())

    def OpenPlugin(self, OpenFrom):
        _MsgItems = [
            self.s2f("Python"),
            self.s2f(""),
            self.s2f("python.cpp: compiling..."),
            self.s2f("   0 error(s), 0 warning(s) :-)"),
            self.s2f(""),
            self.s2f("\x01"),
            self.s2f("&Ok"),
        ]
        MsgItems = self.ffi.new("wchar_t *[]", _MsgItems)
        self.info.Message(
            self.info.ModuleNumber,  # GUID
            self.ffic.FMSG_WARNING | self.ffic.FMSG_LEFTALIGN,  # Flags
            self.s2f("Contents"),  # HelpTopic
            MsgItems,  # Items
            len(MsgItems),  # ItemsNumber
            1,  # ButtonsNumber
        )
        return -1



import logging
import debugpy
from far2l.plugin import PluginBase
from far2l.farprogress import ProgressMessage
from far2l.fardialogbuilder import (
    TEXT,
    EDIT,
    BUTTON,
    MASKED,
    HLine,
    HSizer,
    VSizer,
    DialogBuilder,
)


log = logging.getLogger(__name__)


class Config:
    configured = False
    logto = "/tmp"
    host = "127.0.0.1"
    port = 5678


class Plugin(PluginBase):
    label = "Python udebug"
    openFrom = ["PLUGINSMENU", "COMMANDLINE", "EDITOR", "VIEWER", "FILEPANEL"]

    def debug(self):
        if Config.configured:
            self.parent.Message(("udebug can be configured only once.",))
            return

        t = ProgressMessage(self, self.label, "Waiting for VSCode", 100)
        t.show()
        try:
            Config.configured = True
            debugpy.log_to(Config.logto)
            # in vs code debuger select attach, port = 5678
            # commands in shell:
            #   py:debug
            # elsewhere in python code:
            #   import debugpy
            #   debugpy.breakpoint()
            debugpy.listen((Config.host, Config.port))
            debugpy.wait_for_client()
            # debugpy.breakpoint()
        except:
            log.exception('configured')
        finally:
            t.close()

    def breakpoint(self):
        debugpy.breakpoint()

    @staticmethod
    def HandleCommandLine(line):
        return line in ("debug", "breakpoint")

    def CommandLine(self, line):
        getattr(self, line)()

    def Configure(self):
        if Config.configured:
            self.parent.Message(("udebug can be configured only once.",))
            return

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            if Msg == self.ffic.DN_INITDIALOG:
                try:
                    dlg.SetText(dlg.ID_logpath, Config.logto)
                    dlg.SetText(dlg.ID_ipaddress, Config.host)
                    dlg.SetText(dlg.ID_port, str(Config.port))
                except:
                    log.exception("bang")
            return self.info.DefDlgProc(hDlg, Msg, Param1, Param2)

        b = DialogBuilder(
            self,
            DialogProc,
            "debugpy config",
            "helptopic",
            0,
            VSizer(
                HSizer(
                    TEXT(None, "log path:"),
                    EDIT("logpath", 33, maxlength=80),
                ),
                HSizer(
                    TEXT(None, "ip address:"),
                    EDIT("ipaddress", 20),
                ),
                HSizer(
                    TEXT(None, "ip port:"),
                    MASKED("port", "99999"),
                ),
                HLine(),
                HSizer(
                    BUTTON("vok", "OK", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    BUTTON("vcancel", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        # debugpy.breakpoint()
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == dlg.ID_vok:
            Config.logto = dlg.GetText(dlg.ID_logpath)
            Config.host = dlg.GetText(dlg.ID_ipaddress)
            Config.port = int(dlg.GetText(dlg.ID_port), 10)
        self.info.DialogFree(dlg.hDlg)
        log.debug("end dialog")

    def OpenPlugin(self, OpenFrom):
        self.Configure()



import os
import stat
import time
from datetime import datetime
import io
import subprocess
import threading
import re
import logging


log = logging.getLogger(__name__)


class Entry(object):
    def __init__(
        self,
        dirname=None,
        perms=None,
        links=0,
        uid=0,
        gid=0,
        size=None,
        devmaj=None,
        devmin=None,
        date=None,
        name=None,
        date_re=None,
        datezone=None,
    ):
        self.st_mode = 0
        self.st_perms = perms
        self.st_links = int(links)
        self.st_uid = int(uid)
        self.st_gid = int(gid)
        if devmin is not None:
            self.devmaj = int(devmaj)
            self.devmin = int(devmin)
            self.st_size = 0
        else:
            self.devmaj = None
            self.devmin = None
            self.st_size = int(size)
        self.st_mtime = datetime.strptime(date, date_re)
        self.st_name = name

        if (self.st_perms[0] if self.st_perms else None) == "l" and " -> " in self.st_name:
            try:
                name, target = self.st_name.split(" -> ")
            except ValueError:
                return
            self.st_name = name

        self.perms2mode()

    def perms2mode(self):
        perms = self.st_perms[1:]
        lperms = len(perms)
        mode = 0
        # Check if 'perms' has the right format
        if not [x for x in perms if x not in "-rwxXsStT"] and lperms == 9:
            pos = lperms - 1
            for c in perms:
                if c in "sStT":
                    # Special modes
                    mode += (1 << pos // 3) << 9

                mode += 1 << pos if c in "rwxst" else 0
                pos -= 1
        self.st_mode = (
            mode
            | {
                "-": 0,
                "b": stat.S_IFBLK,
                "c": stat.S_IFCHR,
                "d": stat.S_IFDIR,
                "l": stat.S_IFLNK,
                "p": 0,  # stat.S_IF???,
                "s": 0,  # stat.S_IF???,
            }[self.st_perms[0]]
        )

    def __str__(self):
        if not self.st_name:
            return ""

        template = (
            "{st_mode:5o} {st_perms} {st_links:>4} {st_uid:<8} {st_gid:<8} {st_size:>8} {st_mtime} {st_name}"
        )
        return template.format(**self.__dict__)


class Docker(object):
    def __init__(self, dockerexecutable="/usr/bin/docker"):
        self.dockerexecutable = dockerexecutable

    class Runner(threading.Thread):
        def __init__(self, cmd, stderr=False):
            super().__init__()
            self.cmd = cmd
            self.stderr = stderr
            self.done = threading.Event()
            # 1=killed, 2=exit!=0 3=stderr!=''
            self.error = 0
            self.errors = None
            self.output = None

        def kill(self):
            self.proc.kill()
            self.error = 1

        @property
        def isDone(self):
            return self.done.is_set()

        def run(self):
            self.proc = subprocess.Popen(self.cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            self.error = 0
            try:
                outs, errs = self.proc.communicate()
            except:
                self.proc.kill()
                outs = errs = b""
                self.error = 2
            if errs != b"":
                fp = io.BytesIO(errs)
                self.errors = fp.readlines()
                if not self.stderr:
                    self.error = 3
            else:
                self.errors = []
            if outs != b"":
                fp = io.BytesIO(outs)
                self.output = fp.readlines()
            else:
                self.output = []
            self.done.set()

    def run(self, *args, timeout=2, stderr=False):
        cmd = [self.dockerexecutable]
        cmd.extend(args)
        log.debug(f"docker.run: {cmd}")
        res = subprocess.run(cmd, capture_output=True, timeout=timeout)
        log.debug(f"docker.run: rc={res.returncode}")
        res.check_returncode()
        if stderr:
            fp = io.BytesIO(res.stdout)
            lines1 = fp.readlines()
            fp = io.BytesIO(res.stderr)
            lines2 = fp.readlines()
            return lines1, lines2
        assert res.stderr == b""
        fp = io.BytesIO(res.stdout)
        lines = fp.readlines()
        return lines

    def list(self):
        lines = self.run("container", "ps", "-a")
        devices = []
        if len(lines) > 0 and lines[0].split()[0] == b"CONTAINER":
            line = lines[0].decode()
            cid = 0
            cimage = line.find("IMAGE", 0)
            ccommand = line.find("COMMAND", cimage)
            ccreated = line.find("CREATED", ccommand)
            cstatus = line.find("STATUS", ccreated)
            cports = line.find("PORTS", cstatus)
            cnames = line.find("NAMES", cports)
            for line in lines[1:]:
                line = line.decode()
                info = (line[cid:cimage].strip(), line[cnames:].strip(), line[cstatus:cports].strip().split()[0] == 'Up')
                devices.append(info)
        return devices

    def start(self, name, thread=False):
        if thread:
            t = self.Runner([self.dockerexecutable, "container", "start", name])
            t.start()
            return t
        self.run("container", "start", name)

    def stop(self, name, thread=False):
        if thread:
            t = self.Runner([self.dockerexecutable, "container", "stop", name])
            t.start()
            return t
        self.run("container", "stop", name)

    def logs(self, name):
        return self.run("container", "logs", name, stderr=True)

    date_re = "%Y-%m-%d %H:%M:%S"
    file_re1 = r"""
^
(?P<perms>[\-bcdlps][\-rwxsStT]{9})\s+
(?P<links>\d+)\s+
(?P<uid>\d+)\s+
(?P<gid>\d+)\s+
(?P<devmaj>\d+),\s+(?P<devmin>\d+)\s+
(?P<date>\d{4}-\d{2}-\d{2}\s+\d{2}:\d{2}:\d{2})
(?P<datezone>(\.\d+)?\s+\+\d{4})\s+
(?P<name>.*)
"""
    file_re2 = r'''
^
(?P<perms>[\-bcdlps][\-rwxsStT]{9})\s+
(?P<links>\d+)\s+
(?P<uid>\d+)\s+
(?P<gid>\d+)\s+
(?P<size>\d+)\s+
(?P<date>\d{4}-\d{2}-\d{2}\s+\d{2}:\d{2}:\d{2})
(?P<datezone>(\.\d+)?\s+\+\d{4})\s+
(?P<name>.*)
'''

    def ls(self, deviceid, top):
        file_re1 = re.compile(self.file_re1, re.VERBOSE)
        file_re2 = re.compile(self.file_re2, re.VERBOSE)
        lines = self.run(
            "exec", "-u", "root:root", deviceid, "/bin/ls", "-anL", "--full-time", top
        )
        result = []
        for line in lines:
            line = line.strip()
            if not line:
                continue
            line = line.decode()
            rm = file_re1.match(line)
            if not rm:
                rm = file_re2.match(line)
            if not rm:
                if line.split()[0] != "total":
                    log.debug("ignored entry in ({}): {}".format(top, line))
                continue
            entry = Entry(**rm.groupdict(), date_re=self.date_re)
            result.append(entry)
        return result

    def pull(self, deviceid, sqname, dqname, thread=False):
        if thread:
            t = self.Runner([self.dockerexecutable, "cp", f"{deviceid}:{sqname}", dqname])
            t.start()
            return t
        lines = self.run("cp", f"{deviceid}:{sqname}", dqname)
        for line in lines:
            log.debug(f"pull: {line}")

    def push(self, deviceid, sqname, dqname, thread=False):
        if thread:
            t = self.Runner([self.dockerexecutable, "cp", sqname, f"{deviceid}:{dqname}"])
            t.start()
            return t
        lines = self.run("cp", sqname, f"{deviceid}:{dqname}")
        for line in lines:
            log.debug(f"push: {line}")

    def mkdir(self, deviceid, dqname):
        self.run("exec", "-u", "root:root", deviceid, "mkdir", dqname)

    def remove(self, deviceid, dqname):
        self.run("exec", "-u", "root:root", deviceid, "rm", "-rf", dqname)



import os
import stat
import mmap
import logging
import far2lc
from far2l import fardialogbuilder as fdb
from far2l.plugin import PluginBase
from far2l.far2leditor import Editor

log = logging.getLogger(__name__)

class FileInfo:
    def __init__(self, mmap, filename, filesize):
        self._mmap = mmap
        self._filename = filename
        self._filesize = filesize
        self._fileoffset = 0
        self._modified = {}

    @property
    def mmap(self):
        return self._mmap

    @property
    def filename(self):
        return self._filename

    @property
    def filesize(self):
        return self._filesize

    def get_fileoffset(self):
        return self._fileoffset

    def set_fileoffset(self, fileoffset):
        self._fileoffset = fileoffset

    fileoffset = property(get_fileoffset, set_fileoffset)

    @property
    def modified(self):
        return self._modified

    def modifiy(self, offset, b):
        self._modified[offset] = b


class Screen:
    MIN_WIDTH = 80
    MIN_HEIGHT = 1
    def __init__(self, plugin, width, height):
        self._plugin = plugin
        self._width = max(width, self.MIN_WIDTH)
        self._height = max(height, self.MIN_HEIGHT)
        self._buffer = None

    @property
    def plugin(self):
        return self._plugin

    @property
    def ffi(self):
        return self._plugin.ffi

    @property
    def ffic(self):
        return self._plugin.ffic

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def buffer(self):
        return self._buffer

    def alloc(self, color):
        self._buffer = self.ffi.new("CHAR_INFO []", self._width*self._height)
        self.fillbuffer(color)

    def fillbuffer(self, color):
        for i in range(self._width*self._height):
            self._buffer[i].Char.AsciiChar = b' '
            self._buffer[i].Attributes = color

    def write(self, row, col, val=None, attr=None):
        b = self._buffer
        start = row * self._width + col
        if val is not None:
            for i in range(len(val)):
                b[start + i].Char.UnicodeChar = ord(val[i])
                if attr is not None:
                    b[start + i].Attributes = attr

    def write_bytes(self, row, col, val):
        b = self._buffer
        start = row * self._width + col
        for i in range(len(val)):
            self._buffer[start+i].Char.AsciiChar = bytes([val[i]])

class StatusBar(Screen):
    def __init__(self, plugin, fileinfo, width):
        super().__init__(plugin, width, 1)
        self.color = self.plugin.GetColor(self.ffic.COL_EDITORSTATUS)
        super().alloc(self.color)
        self.positions = [
            0,                    # file name
            self.width-8-1-8-1-1, # modified
            self.width-8-1-8,     # offset
            self.width-8-1,       # file size
        ]
        self._fileinfo = fileinfo
        self.write_filename()
        self.write_filesize()
        self.write_modified()
        self.write_offset(0)

    def write_filename(self):
        maxlen = self.positions[1]
        filename = self._fileinfo.filename
        if len(filename) > maxlen:
            filename = '...'+filename[:maxlen-3]
        self.write(0, 0, filename)

    def write_modified(self):
        self.write(0, self.positions[1], [' ','*'][len(self._fileinfo.modified)!=0])

    def write_offset(self, offset):
        self.write(0, self.positions[2], f'{offset:08x}')

    def write_filesize(self):
        self.write(0, self.positions[3], f'/{self._fileinfo.filesize:08x}')


class KeyBar(Screen):
    ST_NORMAL = 1<<0
    ST_ALT = 1<<1
    ST_SHIFT = 1<<2
    ST_CTRL = 1<<3
    def __init__(self, plugin, fileinfo, width):
        super().__init__(plugin, width, 1)
        self._color_num = self.plugin.GetColor(self.ffic.COL_KEYBARNUM)
        self._color_txt = self.plugin.GetColor(self.ffic.COL_KEYBARTEXT)
        super().alloc(self._color_txt)
        self._state = self.ST_NORMAL
        self._fileinfo = fileinfo
        self.set_labels()

    def get_state(self):
        return self._state

    def set_state(self, state):
        self._state = state
        self.set_labels()

    state = property(get_state, set_state)

    def write_label(self, no, text):
        w = self.width // 12
        col = w * (no-1)
        w -= 2
        if len(text) < w:
            text = text+' '*(w-len(text))
        if len(text) > w:
            text = text[:w]
        self.write(0, col, f'{no:2d}', self._color_num)
        self.write(0, col+2, text)

    def set_labels(self):
        for i in range(12):
            self.write_label(i+1, '')
        if self._state == self.ST_NORMAL:
            self.write_label(1, "Help")
            self.write_label(2, "Save")
            self.write_label(5, "Goto")
            self.write_label(7, "Search")
            self.write_label(9, "Setup")
            self.write_label(10, "Quit")
        elif self._state == self.ST_ALT:
            self.write_label(7, "Prev")
        elif self._state == self.ST_SHIFT:
            self.write_label(2, "SaveAs")
            self.write_label(7, "Next")
        elif self._state == self.ST_CTRL:
            pass

    def col2fkey(self, col):
        w = self.width // 12
        return (col // w)+1


class HexEditor(Screen):
    COL = 12
    def __init__(self, plugin, fileinfo, width, height):
        super().__init__(plugin, width, height)
        self.color = self.plugin.GetColor(self.ffic.COL_EDITORTEXT)
        self.colormodified = 0xe0
        super().alloc(self.color)
        self._fileinfo = fileinfo

    def fill(self):
        super().fillbuffer(self.color)
        fileoffset = self._fileinfo.fileoffset
        filesize = self._fileinfo.filesize
        mmap = self._fileinfo.mmap
        for row in range(self.height):
            self.write(row, 0, f'{fileoffset:010x}:')
            n = 12
            for col in range(16):
                if fileoffset < filesize:
                    b = mmap[fileoffset]
                    self.write(row, n, f'{b:02x}')
                else:
                    self.write(row, n, '  ')
                n += 3
                fileoffset += 1
            b = mmap[fileoffset-16:fileoffset]
            b = bytes([c if c!=0 else 32 for c in b])
            if len(b) < 16:
                b += bytes([32]*(16-len(b)))
            self.write_bytes(row, 64, b)
            if fileoffset >= filesize:
                b = ' '*self.width
                row += 1
                while row < self.height:
                    self.write(row, 0, b)
                    row += 1
                break
        filemin = self._fileinfo.fileoffset
        for k, b in self._fileinfo.modified.items():
            if filemin <= k < fileoffset:
                row = (k - filemin) // 16
                col = ((k - filemin) % 16)
                self.write(row, col * 3 + 12, f'{b:02x}', self.colormodified)
                self.write_bytes(row, 64 + col, bytes([b]))

    def Offset(self, row, col):
        offset = self._fileinfo.fileoffset + row * 16
        if 12 <= col < 60:
            i = (col-12)//3*3+12
            lo = 1 if col > i else 0
            offset += (col-12)//3
            return offset, lo, False
        elif 64 <= col < 80:
            return offset+col-64, 0, True
        return None, None, False

    def Click(self, row, col):
        if 12 <= col < 60:
            i = (col-12)//3*3+12
            i += 1 if col > i else 0
            return i
        elif 64 <= col < 80:
            return col
        return

    def Tab(self, row, col):
        if 12 <= col < 60:
            return row, (col-12)//3+64
        elif 64 <= col < 80:
            return row, 12 + (col - 64) * 3
        return None, None

    def Move(self, row, col, rowinc, colinc):
        if rowinc is not None:
            row += 1 if rowinc else -1
        else:
            if 12 <= col < 60:
                i = (col-12)//3*3+12
                if col == i:
                    i += 1 if colinc else -2
                elif col > i:
                    i += 3 if colinc else 0
                if i == 60:
                    i = 12
                elif i == 10:
                    i = 58
            elif 64 <= col < 80:
                i = col + (1 if colinc else -1)
                if i < 64:
                    i = 79
                elif i == 80:
                    i = 64
            else:
                return None, None
            col = i
        return row, col

class Plugin(PluginBase):
    label = "Python Hex Edit"
    openFrom = ["PLUGINSMENU", "EDITOR", "VIEWER"]

    def OpenPlugin(self, OpenFrom):
        # 5=edit, 6=viewer, 1=panel
        if OpenFrom == 5:
            fqname = Editor(self).GetFileName()
        elif OpenFrom == 6:
            data = self.ffi.new("struct ViewerInfo *")
            data.StructSize = self.ffi.sizeof("struct ViewerInfo")
            self.info.ViewerControl(self.ffic.VCTL_GETINFO, data)
            fqname = self.f2s(data.FileName)
        elif OpenFrom == 1:
            pnli, pnlidata = self.panel.GetCurrentPanelItem()
            fname = self.f2s(pnli.FindData.lpwszFileName)
            fqname = os.path.join(self.panel.GetPanelDir(), fname)
        else:
            log.debug(f"unsupported open from {OpenFrom}")
            return
        # log.debug(f"fqname={fqname}")
        try:
            with open(fqname, "r+b") as f:
                mm = mmap.mmap(f.fileno(), 0)
                try:
                    self.HexEdit(fqname, mm)
                except Exception as ex:
                    log.exception('run')
                finally:
                    mm.close()
        except Exception as ex:
            log.exception('run')

    def GotoDialog(self):
        b = fdb.DialogBuilder(
            self,
            self.info.DefDlgProc,
            "Goto",
            "charmap",
            0,
            fdb.VSizer(
                fdb.HSizer(
                    fdb.TEXT(None, "&Goto:"),
                    fdb.EDIT("vgoto", width=10, history='uhexeditvgoto', flags=self.ffic.DIF_USELASTHISTORY),
                ),
                fdb.HLine(),
                fdb.HSizer(
                    fdb.BUTTON("vok", "OK", default=True, flags=self.ffic.DIF_CENTERGROUP),
                    fdb.BUTTON("vcancel", "Cancel", flags=self.ffic.DIF_CENTERGROUP),
                ),
            ),
        )
        dlg = b.build(-1, -1)

        res = self.info.DialogRun(dlg.hDlg)
        if res == dlg.ID_vok:
            v = dlg.GetText(dlg.ID_vgoto)
            try:
                offset = int(v, 16 if v[:2] == '0x' else 10)
                return offset
            except:
                log.exception('goto')
                return None
        return None

    def GetColor(self, no):
        data = self.ffi.new("DWORD *")
        rc = self.info.AdvControl(
            self.info.ModuleNumber,
            self.ffic.ACTL_GETCOLOR,
            self.ffi.cast("VOID *", no),
            self.ffi.cast("VOID *", data)
        )
        return data[0]

    def GetFarRect(self):
        data = self.ffi.new("SMALL_RECT *")
        rc = self.info.AdvControl(
            self.info.ModuleNumber,
            self.ffic.ACTL_GETFARRECT,
            self.ffi.cast("VOID *", data),
            self.ffi.NULL
        )
        return data

    def Help(self):
        # TODO
        self.notice("TODO: F1=Help.", "HexEditor")

    def Save(self, nname=False, showinfo=True):
        if nname:
            # TODO save as
            self.notice("TODO: F2+SHIFT=Save as ...", "HexEditor")
            return
        else:
            if not self.fileinfo.modified:
                self.notice("File is not modified - not saved.", "HexEditor")
                return
            mm = self.fileinfo.mmap
            for offset, b in self.fileinfo.modified.items():
                mm[offset] = b
        self.fileinfo._modified = {}
        self.hexeditor.fill()
        self.dlg.ShowItem(self.dlg.ID_statusbar, 1)
        self.dlg.ShowItem(self.dlg.ID_hexeditor, 1)
        if showinfo:
            self.notice("File is saved.", "HexEditor")

    def Goto(self):
        offset = self.GotoDialog()
        if offset is None or offset < 0 or offset >= self.fileinfo.filesize:
            return None, None
        if self.fileinfo.fileoffset <= offset and offset < 16 * self.hexeditor.height:
            offset -= self.fileinfo.fileoffset
            row = offset // 16
            col = 12+(offset % 16)*3
        else:
            self.fileinfo.fileoffset = max(offset - 8 * self.hexeditor.height, 0)
            offset -= self.fileinfo.fileoffset
            row = offset // 16
            col = 12+(offset % 16)*3
            self.hexeditor.fill()
        return row, col

    def Search(self):
        # TODO
        self.notice("TODO: F7=Search +ALT=Prev +SHIFT=Next.", "HexEditor")
    SearchNext = SearchPrev = Search

    def Setup(self):
        # TODO
        self.notice("Not implemented F9=Setup.", "HexEditor")

    def Quit(self):
        if not self.fileinfo.modified:
            return 0
        rc = self._popup("File is modified. Save ?", "HexEditor", self.ffic.FMSG_MB_YESNOCANCEL)
        if rc == 0: # Yes
            self.Save(False, True)
            return 0
        elif rc == 1: # NO
            return 0
        return 1

    def Resize(self, size):
        # TODO resize
        pass

    def Scroll(self, size):
        if size < 0:
            fileoffset = max(0, self.fileinfo.fileoffset + size)
            if fileoffset == self.fileinfo.fileoffset:
                return False
        else:
            fileoffset = self.fileinfo.fileoffset + size
            if fileoffset >= self.fileinfo.filesize:
                return False
        self.fileinfo.fileoffset = fileoffset
        self.hexeditor.fill()
        self.dlg.ShowItem(self.dlg.ID_hexeditor, 1)
        return True

    def DialogProc(self, Msg, Param1, Param2):
        # log.debug(f"dlg: Msg={Msg:08x} Param1={Param1:08x} Param2={Param2:08x}")
        if Msg == self.ffic.DN_INITDIALOG:
            self.hexeditor.fill()
            self.dlg.SetCursorPos(self.dlg.ID_hexeditor, self.hexeditor.COL, 0)
            self.dlg.SetCursorSize(self.dlg.ID_hexeditor, 1, 100)
            self.dlg.SetFocus(self.dlg.ID_hexeditor)
            return self.info.DefDlgProc(self.dlg.hDlg, Msg, Param1, Param2)
        elif Msg == self.ffic.DN_RESIZECONSOLE:
            size = self.ffi.cast("COORD *", Param2)
            self.Resize(size)
            return 1
        elif Msg == self.ffic.DN_KEY and Param1 == -1:
            st = KeyBar.ST_NORMAL
            if Param2 & self.ffic.KEY_SHIFT:
                st = KeyBar.ST_SHIFT
            elif Param2 & self.ffic.KEY_CTRL:
                st = KeyBar.ST_CTRL
            elif Param2 & self.ffic.KEY_ALT:
                st = KeyBar.ST_ALT
            if self.keybar.state != st:
                self.keybar.state = st
                self.dlg.ShowItem(self.dlg.ID_keybar, 1)
                return 1
        elif Msg == self.ffic.DN_KEY and Param1 == self.dlg.ID_hexeditor:
            # log.debug(f"dlg.DN_KEY: Param1={Param1:08x} Param2={Param2:08x}")
            col, row = self.dlg.GetCursorPos(self.dlg.ID_hexeditor)
            if Param2 == self.ffic.KEY_PGUP:
                self.Scroll(-16 * self.hexeditor.height)
            elif Param2 == self.ffic.KEY_PGDN:
                self.Scroll(16 * self.hexeditor.height)
            elif Param2 == self.ffic.KEY_LEFT:
                row, col = self.hexeditor.Move(row, col, None, False)
            elif Param2 == self.ffic.KEY_UP:
                row, col = self.hexeditor.Move(row, col, False, None)
                if row == -1:
                    self.Scroll(-16)
                    row = 0
            elif Param2 == self.ffic.KEY_RIGHT:
                row, col = self.hexeditor.Move(row, col, None, True)
                offset, lo, ascii = self.hexeditor.Offset(row, col)
                if offset >= self.fileinfo.filesize:
                    return 1
            elif Param2 == self.ffic.KEY_DOWN:
                row, col = self.hexeditor.Move(row, col, True, None)
                offset, lo, ascii = self.hexeditor.Offset(row, col)
                if offset >= self.fileinfo.filesize:
                    return 1
                if row == self.hexeditor.height:
                    self.Scroll(16)
            elif Param2 == self.ffic.KEY_TAB:
                row, col = self.hexeditor.Tab(row, col)
            elif Param2 == self.ffic.KEY_ESC:
                return self.Quit()
            elif Param2 == self.ffic.KEY_F1:
                self.Help()
                return 1
            elif Param2 in (
                self.ffic.KEY_F2,
                self.ffic.KEY_F2|self.ffic.KEY_SHIFT,
            ):
                self.Save(Param2 == self.ffic.KEY_F2|self.ffic.KEY_SHIFT, True)
                return 1
            elif Param2 == self.ffic.KEY_F5:
                row, col = self.Goto()
                if row is None:
                    return 1
            elif Param2 in (
                self.ffic.KEY_F7, 
                self.ffic.KEY_F7|self.ffic.KEY_SHIFT,
                self.ffic.KEY_F7|self.ffic.KEY_ALT,
            ):
                if Param2 == self.ffic.KEY_F7:
                    self.Search()
                elif Param2 == self.ffic.KEY_F7|self.ffic.KEY_SHIFT:
                    self.SearchNext()
                else:
                    self.SearchPrev()
                return 1
            elif Param2 == self.ffic.KEY_F9:
                self.Setup()
                return 1
            elif Param2 == self.ffic.KEY_F10:
                return self.Quit()
            else:
                if 48 <= Param2 <= 57 or 97 <= Param2 <= 102:
                    offset, lo, ascii = self.hexeditor.Offset(row, col)
                    if not ascii:
                        b = self.fileinfo.modified.get(offset) or self.fileinfo.mmap[offset]
                        v = Param2 - (48 if Param2 < 91 else 87)
                        if lo:
                            b = b&0xf0|v
                        else:
                            b = b&0x0f|(v<<4)
                        self.fileinfo.modifiy(offset, b)
                        self.hexeditor.fill()
                        row, col = self.hexeditor.Move(row, col, None, True)
                else:
                    return self.info.DefDlgProc(self.dlg.hDlg, Msg, Param1, Param2)
            if row is not None:
                self.dlg.SetCursorPos(self.dlg.ID_hexeditor, col, row)
                self.statusbar.write_offset(self.hexeditor.Offset(row, col)[0])
                self.dlg.ShowItem(self.dlg.ID_statusbar, 1)
            return 1
        elif Msg == self.ffic.DN_MOUSECLICK:
            mou = self.ffi.cast("MOUSE_EVENT_RECORD *", Param2)
            col = mou.dwMousePosition.X
            row = mou.dwMousePosition.Y
            if Param1 == self.dlg.ID_hexeditor:
                col = self.hexeditor.Click(row, col)
                if col is not None:
                    self.dlg.SetCursorPos(self.dlg.ID_hexeditor, col, row)
            elif Param1 == self.dlg.ID_keybar:
                fkey = self.keybar.col2fkey(col)
                if self.keybar.state == KeyBar.ST_NORMAL:
                    if fkey == 1:
                        self.Help()
                    elif fkey == 2:
                        self.Save(False, True)
                    elif fkey == 5:
                        row, col = self.Goto()
                        if row is not None:
                            self.dlg.SetCursorPos(self.dlg.ID_hexeditor, col, row)
                    elif fkey == 7:
                        self.Search()
                    elif fkey == 9:
                        self.Setup()
                    elif fkey == 10:
                        if self.Quit() == 0:
                            self.dlg.Close(1)
                elif self.state == KeyBar.ST_ALT:
                    if fkey == 7:
                        self.SearchPrev()
                elif self.state == KeyBar.ST_SHIFT:
                    if fkey == 2:
                        self.Save(True, True)
                    elif fkey == 7: # search next
                        self.SearchNext()
            return 1
        return self.info.DefDlgProc(self.dlg.hDlg, Msg, Param1, Param2)

    def HexEdit(self, fqname, mm):
        # import debugpy; debugpy.breakpoint()

        r = self.GetFarRect()
        self.wwidth = max(r.Right-r.Left+1, 80)
        self.wheight = max(r.Bottom-r.Top+1, 3)
        self.fileinfo = FileInfo(mm, fqname, os.stat(fqname)[stat.ST_SIZE])
        self.hexeditor = HexEditor(self, self.fileinfo, self.wwidth, self.wheight-2)
        self.statusbar = StatusBar(self, self.fileinfo, self.wwidth)
        self.keybar = KeyBar(self, self.fileinfo, self.wwidth)

        @self.ffi.callback("FARWINDOWPROC")
        def DialogProc(hDlg, Msg, Param1, Param2):
            try:
                return self.DialogProc(Msg, Param1, Param2)
            except:
                log.exception('dialogproc')
                return 1

        b = fdb.DialogBuilder(
            self,
            DialogProc,
            "Python HexEdit",
            "hexedit",
            0,
            fdb.VSizer(
                fdb.USERCONTROL('statusbar', self.statusbar.width, self.statusbar.height,
                    param={'VBuf':self.ffi.cast("CHAR_INFO *", self.ffi.addressof(self.statusbar.buffer))}),
                fdb.USERCONTROL('hexeditor', self.hexeditor.width, self.hexeditor.height,
                    param={'VBuf':self.ffi.cast("CHAR_INFO *", self.ffi.addressof(self.hexeditor.buffer))}),
                fdb.USERCONTROL('keybar', self.keybar.width, self.keybar.height,
                    param={'VBuf':self.ffi.cast("CHAR_INFO *", self.ffi.addressof(self.keybar.buffer))}),
            ),
            (0, 0, 0, 0)
        )
        self.dlg = b.build_nobox(-1, -1, self.wwidth, self.wheight)

        self.info.DialogRun(self.dlg.hDlg)
        self.info.DialogFree(self.dlg.hDlg)
        self.dlg = None
        self.keybar = None
        self.statusbar = None
        self.hexeditor = None
        self.fileinfo = None




"""
This plugins allows to quickly jump back and forth between
alpha-numeric words (identifiers). Useful when bound to
hotkeys like Ctrl+Alt+Up / Ctrl+Alt+Down.

Also allows to show all lines with word under cursor.

This is a simple editor plugin which servers and example of
how to make editor plugins using yfar library.
"""

__author__ = 'Yaroslav Yanovsky'

from yfar import FarPlugin


class Plugin(FarPlugin):
    label = 'Jump Between Words in Editor'
    openFrom = ['PLUGINSMENU', 'EDITOR']

    def OpenPlugin(self, _):
        editor = self.get_editor()
        x, y = editor.cursor
        line = editor[y]

        # check if there is a word under cursor
        if x > len(line):
            return

        # extract word
        is_word_char = lambda ch: ch == '_' or ch.isalnum()
        x2 = x
        while x > 0 and is_word_char(line[x - 1]):
            x -= 1
        while x2 < len(line) and is_word_char(line[x2]):
            x2 += 1
        word = line[x:x2]
        if not word:
            return

        def look_back(s, i):
            while i >= 0:
                i = s.rfind(word, 0, i)
                if i != -1:
                    if i > 0 and is_word_char(s[i - 1]):
                        i -= 1
                    else:
                        j = i + len(word)
                        if j < len(s) and is_word_char(s[j]):
                            i -= 1
                        else:
                            return i
            return -1

        def look_forward(s, i):
            while i >= 0:
                i = s.find(word, i)
                if i != -1:
                    if i > 0 and is_word_char(s[i - 1]):
                        i += 1
                    else:
                        j = i + len(word)
                        if j < len(s) and is_word_char(s[j]):
                            i += 1
                        else:
                            return i
            return -1

        option = self.menu(('Jump to &Previous Word', 'Jump to &Next Word',
                            'Display &All lines with Word'),
                           'Jump between words "{}"'.format(word))

        if option == 0:  # move up
            while y >= 0:
                x = look_back(line, x)
                if x == -1:  # not found - search previous line
                    y -= 1
                    if y < 0:
                        break
                    line = editor[y]
                    x = len(line)
                else:
                    editor.focus_cursor(x, y)
                    break

        elif option == 1:  # move down
            total = len(editor)
            x = x2
            while y < total:
                x = look_forward(line, x)
                if x == -1:  # not found - search next line
                    y += 1
                    if y >= total:
                        break
                    line = editor[y]
                    x = 0
                else:
                    editor.focus_cursor(x, y)
                    break

        elif option == 2:  # show all lines with such word
            results = []
            current = 0
            for yi, line in enumerate(editor):
                xi = look_forward(line, 0)
                if xi != -1:
                    if yi == y:
                        current = len(results)
                    results.append((xi, yi, line))
            lno_width = len(str(results[-1][0]))
            fmt = ('{0[1]: >' + str(lno_width) + 'd}:& {0[2]}').format
            i = self.menu([fmt(r) for r in results], word, selected=current)
            if i != -1:
                editor.focus_cursor(results[i][0], results[i][1])
                    
                                



"""
Plugin to rename list of files all at once, using full power of 
FAR text editor.
"""

__author__ = 'Yaroslav Yanovsky'

import os
import tempfile

from yfar import FarPlugin


RENAME_INFO = b'''Rename list of files. File names to the right of slashes
are destination names. File names to the left are given for reference only,
you can modify them too - only the line order is important.
Full destination paths are supported too - either absolute or relative.

'''

class Plugin(FarPlugin):
    label = 'Batch Rename Files and Directories'
    openFrom = ['PLUGINSMENU', 'FILEPANEL']

    def OpenPlugin(self, _):
        panel = self.get_panel()

        dir = panel.directory
        names = []
        for sel in panel.selected:
            full_fn = sel.file_name
            if '/' in full_fn:
                if dir:
                    self.error('Panel directory {} and full panel name '
                               '{} conflict'.format(dir, full_fn))
                    return
            elif dir:
                full_fn = os.path.join(dir, full_fn)
            else:
                self.error('Panel name {} is local and panel directory not '
                           'specified')
                return
            names.append([full_fn])
        if not names:
            self.notice('No files selected')
            return

        max_len = max(len(os.path.basename(x[0])) for x in names)
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(RENAME_INFO)
            for fn, in names:
                fn = os.path.basename(fn)
                f.write(('{} / {}\n'.format(fn.ljust(max_len), fn)).encode())

        renamed = skipped = 0
        failed = []
        if self.editor(f.name, 'Rename list of files', 
                       line=RENAME_INFO.count(b'\n') + 1, column=max_len + 4):
            i = 0
            with open(f.name, 'rb') as f:
                used_names = set()
                for line in f:
                    if b'/' in line:
                        if i >= len(names):
                            self.error('Aborted: edited list has more '
                                       'names than required')
                            break
                        dst_fn = line.split(b'/', 1)[1].strip().decode()
                        if dst_fn.startswith('~'):
                            dst_fn = os.path.expanduser(dst_fn)
                        elif not dst_fn.startswith('/'):
                            dst_fn = os.path.abspath(os.path.join(
                                os.path.dirname(names[i][0]), dst_fn))
                        if dst_fn in used_names:
                            self.error('Aborted: multiple destination names '
                                       'point to the same file\n'
                                       '{}'.format(dst_fn))
                            break
                        names[i].append(dst_fn)
                        used_names.add(dst_fn)
                        i += 1
                else:
                    if i == len(names):
                        for src_fn, dst_fn in names:
                            if src_fn == dst_fn:
                                skipped += 1
                                continue
                            if os.path.exists(dst_fn):
                                failed.append(dst_fn + ': already exists')
                                continue
                            dst_dir, fn = os.path.split(dst_fn)
                            if not os.path.exists(dst_dir):
                                # create dst dir if not exist
                                try:
                                    os.makedirs(dst_dir)
                                except Exception as e:
                                    failed.append(fn + ' - mkdir: ' + str(e))
                                    continue
                            try:
                                os.rename(src_fn, dst_fn)
                            except Exception as e:
                                failed.append(fn + ': ' + str(e))
                            else:
                                renamed += 1
                    else:
                        self.error('Aborted: edited list has less names than '
                                   'required')
            panel.refresh()
            if renamed or skipped or failed:
                msg = '{:d} file(s) renamed'.format(renamed)
                if skipped:
                    msg += ', {:d} skipped'.format(skipped)
                if failed:
                    self.error(msg + ', {:d} failed:\n{}'.format(
                        len(failed), '\n'.join(failed)))
                else:
                    self.notice(msg)

        if os.path.exists(f.name):
            os.unlink(f.name)
                